"""An experiment that enables ``Message.__eq__``.

It compares the IDs of each ``Message``.

Example:

```py
>>> m1 == m1
True # Same IDs
>>> m1 == m2
False # Different IDs
```
"""

import nextcord


nextcord.Message.__eq__ = lambda s, o: isinstance(o, nextcord.Message) and s.id == o.id



import inspect
import typing
import importlib

import nextcord
from nextcord.ext.commands import converter, Context
from .converter import CustomConverter


BASE_CONVERTERS = {
    nextcord.CategoryChannel: converter.CategoryChannelConverter,
    nextcord.Colour:          converter.ColourConverter,
    nextcord.Emoji:           converter.EmojiConverter,
    nextcord.Game:            converter.GameConverter,
    nextcord.Invite:          converter.InviteConverter,
    nextcord.Member:          converter.MemberConverter,
    nextcord.Message:         converter.MessageConverter,
    nextcord.PartialEmoji:    converter.PartialEmojiConverter,
    nextcord.Role:            converter.RoleConverter,
    nextcord.TextChannel:     converter.TextChannelConverter,
    nextcord.User:            converter.UserConverter,
    nextcord.VoiceChannel:    converter.VoiceChannelConverter,
}


class ConverterDict(dict):
    def __init__(self):
        super().__init__(BASE_CONVERTERS)
    
    def __setitem__(self, k: typing.Any, v: typing.Any):
        if not (issubclass(v, converter.Converter) or inspect.isbuiltin(v)):
            raise TypeError("Excepted value of type 'Converter' or built-in, received {}".format(v.__name__))
        super().__setitem__(k, v)
    
    def set(self, _type: typing.Any, converter: str) -> 'ConverterDict':
        self.__setitem__(_type, converter)
        return self
    
    def get(self, _type: typing.Any, default: typing.Any = None) -> 'ConverterDict':
        if inspect.isclass(_type):
            return super().get(_type, default)
        else:
            _converter = super().get(type(_type), default)

            if issubclass(_converter, CustomConverter):
                _converter = _converter[_type]
            
            return _converter
    
    def register(self, _type: type) -> converter.Converter:
        def predicate(handler: typing.Union[typing.Callable[[Context, str], typing.Awaitable[typing.Any]], converter.Converter]):
            # _type is what is should convert to.

            _converter = handler

            if inspect.iscoroutinefunction(handler):
                _converter = type(
                        '{}Converter'.format(_type.__name__),
                        (converter.Converter,),
                        {
                            'convert': lambda s, c, a: handler(c, a) # should mean it's a coroutine.
                        }
                    )
                
                self.set(_type, _converter)
            else:
                self.set(_type, _converter)
            
            return _converter
        return predicate

    def load(self, *_converters: str) -> None:
        for c in _converters:
            importlib.import_module('nextcord.ext.converters.custom_converters.{}'.format(c)).setup(self)



import nextcord

from ._events import _ALL
from .utils import fetch_recent_audit_log_entry, listens_for


EVENT_NAME = 'member_kick'


@listens_for('member_remove')
async def check_member_kick(client: nextcord.Client, member: nextcord.Member):
    guild = member.guild

    print('!RAN')

    if not guild.me.guild_permissions.view_audit_log:
        return

    entry = await fetch_recent_audit_log_entry(client, member.guild, target=member, action=nextcord.AuditLogAction.kick, retry=3)
    if entry is None:
        return

    client.dispatch(EVENT_NAME, member, entry)


_ALL[EVENT_NAME] = check_member_kick



from typing import List, Callable, Tuple
from dataclasses import dataclass

from .flags import OpFlags

OP_DICT: List[Callable[[OpFlags], bool]]= [
    lambda opflag: opflag.DISPATCH,
    lambda opflag: opflag.HEARTBEAT,
    lambda opflag: opflag.IDENTIFY,
    lambda opflag: opflag.PRESENCE,
    lambda opflag: opflag.VOICE_STATE,
    lambda opflag: opflag.VOICE_PING,
    lambda opflag: opflag.RESUME,
    lambda opflag: opflag.RECONNECT,
    lambda opflag: opflag.REQUEST_MEMBERS,
    lambda opflag: opflag.INVALIDATE_MEMBERS,
    lambda opflag: opflag.HELLO,
    lambda opflag: opflag.HEARTBEAT_ACK,
    lambda opflag: opflag.GUILD_SYNC,
]

@dataclass
class OpDetails:
    inbound: bool

    # Op 0
    event_name: str = ""

    payload: dict = None



import time
import sys
from urllib.parse import quote as _uriquote
import weakref
import asyncio

from aiohttp import ClientSession
from nextcord.http import HTTPClient, Route, MaybeUnlock, json_or_text
from nextcord import utils
from nextcord.errors import HTTPException, Forbidden, NotFound, LoginFailure, GatewayNotFound

from .snoopy import Analyst


class Beaker:
    def __init__(self, *args, **kwargs):
        self._pipette: Pipette = kwargs.pop('pipette')
        self._session = ClientSession(*args, **kwargs)
        self.prepare_request_method()
    
    def __getattr__(self, name):
        return getattr(self._session, name)
    
    def prepare_request_method(self):
        _old_request = self._session._request
        async def _request(s, *args, **kwargs):
            start = time.monotonic()
            try:
                ret = await _old_request(s, *args, **kwargs)
            except Exception as e:
                await self._pipette.analyst.http_request_error(*args, kwargs, exc=e)
                raise
            else:
                end = time.monotonic()
                dur = end - start
                await self._pipette.analyst.http_request(*args, kwargs, ret=ret, dur=dur)
                return ret
        self._session._request = _request


class Pipette(HTTPClient):
    session_cls: Beaker
    analyst: Analyst

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        from . import __version__ # circular import
        self.user_agent += f' nextcord-ext-science/{__version__}'
    
    @classmethod
    def from_http_client(cls, http_client: HTTPClient, session_cls: Beaker, analyst: Analyst):
        old = http_client
        self = cls(
            old.connector,
            proxy=old.proxy,
            proxy_auth=old.proxy_auth,
            loop=old.loop,
            unsync_clock=not old.use_clock,
        )
        self.session_cls = session_cls
        self.analyst = analyst
        return self
        
    def recreate(self):
        if self._HTTPClient__session.closed:
            self._HTTPClient__session = self.session_cls(connector=self.connector, pipette=self)

    async def static_login(self, token, *, bot):
        print(dir(self))
        # Necessary to get aiohttp to stop complaining about session creation
        self._HTTPClient__session = self.session_cls(connector=self.connector, pipette=self)
        print(self._HTTPClient__session)
        old_token, old_bot = self.token, self.bot_token
        self._token(token, bot=bot)

        try:
            print(self._HTTPClient__session)
            data = await self.request(Route('GET', '/users/@me'))
        except HTTPException as exc:
            self._token(old_token, bot=old_bot)
            if exc.response.status == 401:
                raise LoginFailure('Improper token has been passed.') from exc
            raise

        return data
# 
# create table 
# requests(
#   id serial primary key, 
#   time time, 
#   verb text, 
#   url text, 
#   data text, 
#   headers text, 
#   params text,
#   result text,
#   error text
# )
#



import logging
from typing import Dict, Union

import nextcord


# consistency with the `nextcord` namespaced logging
log = logging.getLogger(__name__)

# default timeout parameter for menus in seconds
DEFAULT_TIMEOUT = 180.0

# type definition for the keyword-arguments that are
# used in both Message.edit and Messageable.send
SendKwargsType = Dict[str, Union[str, nextcord.Embed, nextcord.ui.View, None]]

# type definition for possible page formats
PageFormatType = Union[str, nextcord.Embed, SendKwargsType]

# type definition for emoji parameters
EmojiType = Union[str, nextcord.Emoji, nextcord.PartialEmoji]



"""An experiment that allows you to define commands in the bot subclass

Example:
```py
from nextcord.ext.alternatives import inline_bot_commands
from nextcord.ext import commands

class MyBot(commands.Bot):
    @commands.command()
    async def test(self, ctx):
        await ctx.send(f'{len(self.all_commands)} commands registered on {self.__class__.__name__}!')

    @commands.command()
    async def echo(self, ctx, *, words):
        await ctx.send(words)


bot = MyBot(command_prefix='?')
bot.run(token)
```
"""

from nextcord.ext import commands


class InlineMeta(type):
    def __new__(cls, *args, **kwargs):
        new_cls = super().__new__(cls, *args)
        cmds = {}
        for base in reversed(new_cls.__mro__):
            for elem, value in base.__dict__.items():
                if elem in cmds:
                    del cmds[elem]
                if isinstance(value, commands.Command):
                    cmds[elem] = value

        new_cls.__inline_commands__ = list(cmds.values())
        return new_cls

    @property
    def qualified_name(cls):
        # for the default help command, since the bot is acting as a cog
        return "No Category"


class BotBase(commands.bot.BotBase, metaclass=InlineMeta):
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)

        self.__inline_commands__ = tuple(c.copy() for c in cls.__inline_commands__)

        lookup = {cmd.qualified_name: cmd for cmd in self.__inline_commands__}

        # Update the Command instances dynamically as well
        for command in self.__inline_commands__:
            setattr(self, command.callback.__name__, command)
            command.cog = self
            parent = command.parent
            if parent is not None:
                # Get the latest parent reference
                parent = lookup[parent.qualified_name]

                # Update our parent's reference to our self
                parent.remove_command(command.name)
                parent.add_command(command)

        return self

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for command in self.__inline_commands__:
            self.add_command(command)


commands.bot.BotBase = BotBase



"""An experiment that allows handlers to handle specific errors.

This overrides the default behaviour of ``Command.error``.
(Change ``@Command.error`` to ``@Command.error()``).

Example:
```py
@bot.command()
async def throw_error(ctx):
    raise Exception

@command.error(CommandInvokeError)
async def handle(ctx, error):
    if isinstance(error.original, Exception):
        await ctx.send('oh no')
```
"""

import asyncio

from nextcord.ext import commands


async def on_error(*args):  # cog?, ctx, error
    error, ctx = args[-1], args[-2]
    cls = error.__class__

    for exc, callback in ctx.command._handled_errors.items():
        if exc in cls.__mro__:  # walk mro to check if the handler can handle it
            await callback(*args)


def error(self, *exceptions):
    def decorator(func):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("The error must be a coroutine.")

        if not exceptions:
            self.on_error = func
            return

        try:
            self._handled_errors
        except AttributeError:
            self._handled_errors = {}
        finally:
            for exc in exceptions:
                self._handled_errors[exc] = func

        try:
            self.on_error
        except AttributeError as e:
            self.on_error = on_error

        return func

    return decorator


commands.Command.error = error

_old_ensure_assignment_on_copy = commands.Command._ensure_assignment_on_copy


def _ensure_assignment_on_copy(self, other):
    other = _old_ensure_assignment_on_copy(self, other)

    try:
        other._handled_errors = self._handled_errors
    except AttributeError:
        pass

    return other


commands.Command._ensure_assignment_on_copy = _ensure_assignment_on_copy



from nextcord.colour import Colour


MDIO_COLOURS = {
    "400": {
        "red": 0xEF534E,
        "pink": 0xEC407E,
        "purple": 0xAB47BC,
        "deep_purple": 0x7E56C1,
        "indigo": 0x5C6BC0,
        "blue": 0x42A5F5,
        "light_blue": 0x29B6F6,
        "cyan": 0x26C6DA,
        "teal": 0x26A69A,
        "green": 0x66BB6A,
        "light_green": 0x9CCC65,
        "lime": 0xD4E157,
        "yellow": 0xFFEE58,
        "amber": 0xFFCA28,
        "orange": 0xFFA726,
        "deep_orange": 0xFF7043,
    }
}

for shade, colours in MDIO_COLOURS.items():
    for name, value in colours.items():
        delegate = lambda cls, value=value: cls(value)
        setattr(Colour, "material_{0}_{1}".format(shade, name), classmethod(delegate))



"""
    Copyright (c) 2020 Ext-Creators

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from nextcord.channel import _channel_factory
from nextcord.enums import ChannelType
from nextcord.object import Object
from nextcord.errors import InvalidData

class GetchManager:
    def __init__(self, *, bot):
        self.bot = bot
        self.http = bot.http
        self._connection = bot._connection
    
    def fetch_guild(self, id: int):
        return self.bot.fetch_guild(id)
    
    def get_guild(self, id: int):
        return self.bot.get_guild(id)
    
    async def fetch_channel(self, id: int, *, guild=None):
        data = await self.http.get_channel(id)

        factory, ch_type = _channel_factory(data['type'])
        if factory is None:
            raise InvalidData('Unknown channel type {type} for channel ID {id}.'.format_map(data))

        if ch_type in (ChannelType.group, ChannelType.private):
            channel = factory(me=self.user, data=data, state=self._connection)
        else:
            guild_id = int(data['guild_id'])
            guild = guild or self.get_guild(guild_id) or Object(id=guild_id)
            channel = factory(guild=guild, state=self._connection, data=data)

        return channel
    
    def get_channel(self, id: int):
        return self.bot.get_channel(id)
    
    async def fetch_message(self, id: int, *, channel=None):
        data = await self.http.get_message(channel.id, id)
        return self._connection.create_message(channel=channel, data=data)

    def get_message(self, id: int):
        return self._connection._get_message(id)



"""An experiment that allows you to set a command suffix on a Bot

Example:
```py
bot = commands.Bot(command_suffix='!', command_prefix='!')  # both the prefix and the suffix will work when invoking commands


@bot.command()
async def say(ctx, *, words):
    ret = 'You said \"{}\" with the {} `{}`!'

    if ctx.prefix is not None:
        ret = ret.format(words, 'prefix', ctx.prefix)
    else:
        ret = ret.format(words, 'suffix', ctx.prefix)

    await ctx.send(ret)


@bot.command()
async def ping(ctx):
    await ctx.send('Pong! {}ms'.format(ctx.bot.latency * 1000))


@bot.command()
async def hello(ctx):
    await ctx.send('Hello! I\'m a bot with command suffixes!')
```

'hello!' --> 'Hello! I'm a bot with command suffixes!'
'say! this is neat' --> 'You said "this is neat" with the suffix `!`'
'!say this is neat' --> 'You said "this is neat" with the prefix `!`'
'!say! this is neat' --> No reply
"""

import collections

import nextcord
from nextcord.ext import commands


def _suffix_used(suffix, content):
    space_index = content.find(" ")
    suffix_index = content.find(suffix)
    return suffix_index > 0 and (space_index == -1 or suffix_index < space_index)


class Context(commands.Context):
    def __init__(self, **attrs):
        super().__init__(**attrs)
        self.suffix = attrs.pop("suffix")

    @property
    def valid(self):
        return (self.suffix is not None or self.prefix is not None) and self.command is not None

    async def reinvoke(self, *, call_hooks=False, restart=True):
        if self.suffix is not None:
            # since the command was invoked with a suffix,
            # we need to make sure the view doesn't try to skip a nonexistent prefix
            original_prefix = self.prefix
            self.prefix = ""

        await super().reinvoke(call_hooks=call_hooks, restart=restart)

        try:
            self.prefix = original_prefix
        except NameError:
            pass


class BotBase(commands.bot.BotBase):
    def __init__(self, command_prefix=None, command_suffix=None, **options):
        if command_prefix is None and command_suffix is None:
            raise ValueError("Bot must have a prefix or suffix")

        super().__init__(command_prefix=command_prefix, **options)
        self.command_suffix = command_suffix

    async def get_prefix(self, message):
        if self.command_prefix is None:
            return None

        return await super().get_prefix(message)

    async def get_suffix(self, message):
        """|coro|
        Retrieves the prefix the bot is listening to
        with the message as a context.
        Parameters
        -----------
        message: :class:`nextcord.Message`
            The message context to get the prefix of.
        Returns
        --------
        Optional[Union[List[:class:`str`], :class:`str`]]
            A list of prefixes or a single prefix that the bot is
            listening for.
        """
        if self.command_suffix is None:
            return None

        suffix = ret = self.command_suffix
        if callable(suffix):
            ret = await nextcord.utils.maybe_coroutine(suffix, self, message)

        if not isinstance(ret, str):
            try:
                ret = list(ret)
            except TypeError:
                # It's possible that a generator raised this exception.  Don't
                # replace it with our own error if that's the case.
                if isinstance(ret, collections.abc.Iterable):
                    raise

                raise TypeError(
                    "command_suffix must be plain string, iterable of strings, or callable "
                    "returning either of these, not {}".format(ret.__class__.__name__)
                )

            if not ret:
                raise ValueError("Iterable command_prefix must contain at least one suffix")

        return ret

    async def get_context(self, message, *, cls=Context):
        """Defaults to check for prefix first."""
        view = commands.view.StringView(message.content)
        ctx = cls(prefix=None, suffix=None, view=view, bot=self, message=message)

        if self._skip_check(message.author.id, self.user.id):
            return ctx

        prefix = await self.get_prefix(message)
        suffix = await self.get_suffix(message)

        if prefix is not None:
            if isinstance(prefix, str):
                if view.skip_string(prefix):
                    invoked_prefix = prefix
            else:
                try:
                    if message.content.startswith(tuple(prefix)):
                        invoked_prefix = nextcord.utils.find(view.skip_string, prefix)
                except TypeError:
                    if not isinstance(prefix, list):
                        raise TypeError(
                            "get_prefix must return either a string or a list of string, "
                            "not {}".format(prefix.__class__.__name__)
                        )

                    for value in prefix:
                        if not isinstance(value, str):
                            raise TypeError(
                                "Iterable command_prefix or list returned from get_prefix must "
                                "contain only strings, not {}".format(value.__class__.__name__)
                            )

                    raise
        else:
            if isinstance(suffix, str):
                if _suffix_used(suffix, message.content):
                    invoked_suffix = suffix
                else:
                    return ctx
            else:
                try:
                    invoked_suffixes = [s for s in suffix if _suffix_used(s, message.content)]
                    if not invoked_suffixes:
                        return ctx

                    for suf in invoked_suffixes:
                        invoker = view.get_word()[: -len(suf)]
                        command = self.all_commands.get(invoker)
                        if command is not None:
                            view.undo()
                            invoked_suffix = suf
                            break
                    else:
                        return ctx

                except TypeError:
                    if not isinstance(suffix, list):
                        raise TypeError(
                            "get_suffix must return either a string or a list of string, "
                            "not {}".format(suffix.__class__.__name__)
                        )

                    for value in suffix:
                        if not isinstance(value, str):
                            raise TypeError(
                                "Iterable command_suffix or list returned from get_suffix must "
                                "contain only strings, not {}".format(value.__class__.__name__)
                            )

                    raise

        invoker = view.get_word()

        try:
            ctx.suffix = invoked_suffix
        except NameError:
            try:
                ctx.prefix = invoked_prefix
            except NameError:
                pass
        else:
            invoker = invoker[: -len(invoked_suffix)]

        ctx.invoked_with = invoker
        ctx.command = self.all_commands.get(invoker)
        return ctx


commands.bot.BotBase = BotBase
commands.Context = Context



class MenuError(Exception):
    pass


class CannotEmbedLinks(MenuError):
    def __init__(self):
        super().__init__("Bot does not have embed links permission in this channel.")


class CannotSendMessages(MenuError):
    def __init__(self):
        super().__init__("Bot cannot send messages in this channel.")


class CannotAddReactions(MenuError):
    def __init__(self):
        super().__init__("Bot cannot add reactions in this channel.")


class CannotReadMessageHistory(MenuError):
    def __init__(self):
        super().__init__("Bot does not have Read Message History permissions in this channel.")



"""An experiment that allows for conversion of ``Asset`` 
arguments for commands.

It first detects if there's an attachment available or a URL in the 
message content, it will return it as an ``Asset``. If not, 
it will default onto a "null" ``Asset`` (one with no URL).

Example:
```py
@commands.command()
async def test(ctx, image: Asset):
    asset_bytes = io.BytesIO()
    await image.save(asset_bytes)
    do_some_cool_image_manipulation(asset_bytes)

    await ctx.send(file=nextcord.File(asset_bytes, 'cool_image.png'))
```
"""

from nextcord.ext.commands import converter, Context, errors, Command
from inspect import Parameter
from nextcord import Asset, nextcordException
import typing

from ._common import _ALL

# Basic Asset Converter


class _AssetConverter(converter.Converter):
    async def convert(self, ctx: Context, argument: str):
        if argument.startswith("http"):
            return Asset(ctx.bot._connection, argument)

        raise errors.BadArgument("No image found!")


converter.AssetConverter = _AssetConverter

_ALL[Asset] = _AssetConverter

Asset.__str__ = (
    lambda s: "" if s._url is None else (s._url if s._url.startswith("http") else s.BASE + s._url)
)


async def _read(self):
    if not self._url:
        raise nextcordException("Invalid asset (no URL provided)")

    if self._state is None:
        raise nextcordException("Invalid state (no ConnectionState provided)")

    return await self._state.http.get_from_cdn(str(self))


Asset.read = _read

# "Hijack" transform to set default for Asset to preprocess possibility of attachment

_old_transform = Command.transform


def _transform(self, ctx, param):
    if param.annotation is Asset and param.default is param.empty:
        if ctx.message.attachments:
            default = Asset(ctx.bot._connection, ctx.message.attachments[0].url)
            param = Parameter(
                param.name,
                param.kind,
                default=default,
                annotation=typing.Optional[param.annotation],
            )
        else:
            default = Asset(ctx.bot._connection, "")
            param = Parameter(param.name, param.kind, default=default, annotation=param.annotation)

    return _old_transform(self, ctx, param)


Command.transform = _transform



from nextcord.ext.commands.bot import BotBase, Bot, AutoShardedBot
from .core import ConvertersGroupMixin


class ConvertersBotBase(BotBase, ConvertersGroupMixin):
    ...

class ConvertersBot(Bot, ConvertersBotBase):
    ...

class AutoShardedConvertersBot(AutoShardedBot, ConvertersBotBase):
    ...



"""An experiment that allows you to change the positions of channels
in a CategoryChannel in a way that isn't quickly rate-limited.

Works with TextChannels and VoiceChannels alike.

Example:
```py
@is_owner()
@bot.command()
async def by_length(ctx):
    await ctx.channel.category.sort(key=lambda c: len(c.name))

@is_owner()
@bot.command()
async def alphabetize(ctx):
    await ctx.channel.category.alphabetize()

@is_owner()
@bot.command()
async def shuffle(ctx):
    await ctx.channel.category.shuffle()
```
"""

import random

from nextcord import CategoryChannel


async def _sort(self, *, key=None, reverse=False):
    """|coro|

    Sorts the channels within the CategoryChannel, similar to Python's list.sort().

    You must have the :attr:`~nextcord.Permissions.manage_channels` permission to
    do this.

    Parameters
    -----------
    key: Callable
        A callable function to customize the sort order.
        The supplied argument is of type ``GuildChannel``.
    reverse: :class:`bool`
        Whether or not to sort in descending order. False by default.

    Raises
    -------
    Forbidden
        You do not have permissions to sort the channels.
    HTTPException
        Sorting the channels failed.
    """
    payload = [
        {"id": channel.id, "position": index}
        for index, channel in enumerate(sorted(self.channels, key=key, reverse=reverse))
    ]

    await self._state.http.bulk_channel_update(self.guild.id, payload)


async def _alphabetize(self, *, reverse=False):
    """|coro|

    Alphabetizes the channels within the CategoryChannel.

    You must have the :attr:`~nextcord.Permissions.manage_channels` permission to
    do this.

    Parameters
    -----------
    reverse: :class:`bool`
        Whether or not to alphabetize in descending order. False by default.

    Raises
    -------
    Forbidden
        You do not have permissions to alphabetize the channels.
    HTTPException
        Alphabetizing the channels failed.
    """

    await self.sort(key=lambda c: c.name, reverse=reverse)


async def _shuffle(self):
    """|coro|

    Shuffles the channels within the CategoryChannel.

    You must have the :attr:`~nextcord.Permissions.manage_channels` permission to
    do this.

    Raises
    -------
    Forbidden
        You do not have permissions to shuffle the channels.
    HTTPException
        Shuffling the channels failed.
    """

    await self.sort(key=lambda _: random.random())


CategoryChannel.sort = _sort
CategoryChannel.alphabetise = _alphabetize
CategoryChannel.alphabetize = _alphabetize
CategoryChannel.shuffle = _shuffle



"""
See https://github.com/Rapptz/nextcord.py/issues/4098.
"""
import asyncio
from nextcord import Role, ChannelType, InvalidArgument, PermissionOverwrite
import nextcord.abc


async def _edit(self, options, reason):
    try:
        parent = options.pop("category")
    except KeyError:
        parent_id = nextcord.abc._undefined
    else:
        parent_id = parent and parent.id
    try:
        options["rate_limit_per_user"] = options.pop("slowmode_delay")
    except KeyError:
        pass
    lock_permissions = options.pop("sync_permissions", False)
    try:
        position = options.pop("position")
    except KeyError:
        if parent_id is not nextcord.abc._undefined:
            if lock_permissions:
                category = self.guild.get_channel(parent_id)
                options["permission_overwrites"] = [c._asdict() for c in category._overwrites]
            options["parent_id"] = parent_id
        elif lock_permissions and self.category_id is not None:
            # if we're syncing permissions on a pre-existing channel category without changing it
            # we need to update the permissions to point to the pre-existing category
            category = self.guild.get_channel(self.category_id)
            options["permission_overwrites"] = [c._asdict() for c in category._overwrites]
    else:
        await self._move(
            position, parent_id=parent_id, lock_permissions=lock_permissions, reason=reason
        )
    overwrites = options.get("overwrites", None)
    if overwrites is not None:
        perms = []
        for target, perm in overwrites.items():
            if not isinstance(perm, PermissionOverwrite):
                raise InvalidArgument(
                    "Expected PermissionOverwrite received {0.__name__}".format(type(perm))
                )
            allow, deny = perm.pair()
            payload = {
                "allow": allow.value,
                "deny": deny.value,
                "id": target.id,
            }
            if isinstance(target, Role):
                payload["type"] = "role"
            else:
                payload["type"] = "member"
            perms.append(payload)
        options["permission_overwrites"] = perms
    try:
        ch_type = options["type"]
    except KeyError:
        pass
    else:
        if not isinstance(ch_type, ChannelType):
            raise InvalidArgument("type field must be of type ChannelType")
        options["type"] = ch_type.value

    if options:
        data = await self._state.http.edit_channel(self.id, reason=reason, **options)
        # see issue Rapptz/nextcord.py#4098
        if "parent_id" in options:
            client = self._state._get_client()
            try:
                await client.wait_for(
                    "guild_channel_update",
                    check=lambda b, a: b.id == a.id and b.id == self.id,
                    timeout=2,
                )
                return
            except asyncio.TimeoutError:
                # fallback, we didn't receive the event within 2s
                pass
        self._update(self.guild, data)


nextcord.abc.GuildChannel._edit = _edit



import nextcord
from rx.subject import Subject


class RxBot(nextcord.Client):
    messages = Subject()
    edited_messages = Subject()
    deleted_messages = Subject()

    async def on_message(self, message):
        self.messages.on_next(message)

    async def on_message_delete(self, message):
        self.deleted_messages.on_next(message)
    
    async def on_message_edit(self, before, after):
        self.edited_messages.on_next((before, after))
    
    reactions = Subject()
    removed_reactions = Subject()
    cleared_reactions = Subject()
    cleared_emoji_reactions = Subject()

    async def on_reaction_add(self, reaction, user):
        self.reactions.on_next((reaction, user))
    
    async def on_reaction_remove(self, reaction, user):
        self.removed_reactions.on_next((reaction, user))
    
    async def on_reaction_clear(self, reaction, user):
        self.cleared_reactions.on_next((reaction, user))
    
    async def on_reaction_clear_emoji(self, reaction):
        self.cleared_emoji_reactions.on_next(reaction)



"""An experiment to allow for webhooks to change channels.

Example:
```py
webhook = await ctx.bot.fetch_webhook(id)
channel = ctx.guild.text_channels[1]

await webhook.move_to(channel=channel)
```
"""
from nextcord import utils, errors, webhook, TextChannel
import asyncio
import json
import aiohttp

# register endpoint

_old_prepare = webhook.WebhookAdapter._prepare


def _prepare(self, webhook):
    _old_prepare(self, webhook)
    self._move_url = "{0.BASE}/webhooks/{1}".format(self, webhook.id)


webhook.WebhookAdapter._prepare = _prepare


def _move_webhook(self, token, **payload):
    return self.request("PATCH", self._move_url, payload=payload, token=token)


webhook.WebhookAdapter.move_webhook = _move_webhook

# fix up AsyncWebhookAdapter.request to provide token -- was hoping to not need to replace it completely


async def _request(self, verb, url, payload=None, multipart=None, *, files=None, token=None):
    headers = {}
    if token:
        headers["Authorization"] = token

    data = None
    files = files or []
    if payload:
        headers["Content-Type"] = "application/json"
        data = utils.to_json(payload)

    if multipart:
        data = aiohttp.FormData()
        for key, value in multipart.items():
            if key.startswith("file"):
                data.add_field(key, value[1], filename=value[0], content_type=value[2])
            else:
                data.add_field(key, value)

    for tries in range(5):
        for file in files:
            file.reset(seek=tries)

        async with self.session.request(verb, url, headers=headers, data=data) as r:
            # Coerce empty strings to return None for hygiene purposes
            response = (await r.text(encoding="utf-8")) or None
            if r.headers["Content-Type"] == "application/json":
                response = json.loads(response)

            # check if we have rate limit header information
            remaining = r.headers.get("X-Ratelimit-Remaining")
            if remaining == "0" and r.status != 429:
                delta = utils._parse_ratelimit_header(r)
                await asyncio.sleep(delta)

            if 300 > r.status >= 200:
                return response

            # we are being rate limited
            if r.status == 429:
                retry_after = response["retry_after"] / 1000.0
                await asyncio.sleep(retry_after)
                continue

            if r.status in (500, 502):
                await asyncio.sleep(1 + tries * 2)
                continue

            if r.status == 403:
                raise errors.Forbidden(r, response)
            elif r.status == 404:
                raise errors.NotFound(r, response)
            else:
                raise errors.HTTPException(r, response)
    # no more retries
    raise errors.HTTPException(r, response)


webhook.AsyncWebhookAdapter.request = _request

# add the method + _async_move


async def _async_move(self, channel_id, token, payload):
    await self._adapter.move_webhook(token, **payload)

    self.channel_id = channel_id


def _move_to(self, channel):
    if not isinstance(channel, TextChannel):
        raise TypeError(
            "Expected TextChannel parameter, received %s instead." % channel.__class__.__name__
        )

    payload = {"channel_id": channel.id}

    token = channel._state.http.token

    if channel._state.is_bot:
        token = "Bot " + token

    if isinstance(self._adapter, webhook.AsyncWebhookAdapter):
        return self._async_move(channel.id, token, payload)


webhook.Webhook._async_move = _async_move
webhook.Webhook.move_to = _move_to



from typing import Any, List, Optional

import nextcord
from nextcord.ext import commands

from .constants import PageFormatType, SendKwargsType
from .menus import Button, ButtonMenu, Menu
from .page_source import PageSource
from .utils import First, Last, _cast_emoji


class MenuPagesBase(Menu):
    """A base class dedicated to pagination for reaction and button menus.

    Attributes
    ------------
    current_page: :class:`int`
        The current page that we are in. Zero-indexed
        between [0, :attr:`PageSource.max_pages`).
    """
    FIRST_PAGE = '\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\ufe0f'
    PREVIOUS_PAGE = '\N{BLACK LEFT-POINTING TRIANGLE}\ufe0f'
    NEXT_PAGE = '\N{BLACK RIGHT-POINTING TRIANGLE}\ufe0f'
    LAST_PAGE = '\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\ufe0f'
    STOP = '\N{BLACK SQUARE FOR STOP}\ufe0f'

    def __init__(self, source: PageSource, **kwargs):
        self._source = source
        self.current_page = 0
        if hasattr(self, "__button_menu_pages__"):
            ButtonMenu.__init__(self, **kwargs)
            return
        Menu.__init__(self, **kwargs)

    @property
    def source(self) -> PageSource:
        """:class:`PageSource`: The source where the data comes from."""
        return self._source

    async def change_source(self, source: PageSource):
        """|coro|

        Changes the :class:`PageSource` to a different one at runtime.

        Once the change has been set, the menu is moved to the first
        page of the new source if it was started. This effectively
        changes the :attr:`current_page` to 0.

        Raises
        --------
        TypeError
            A :class:`PageSource` was not passed.
        """

        if not isinstance(source, PageSource):
            raise TypeError(
                'Expected {0!r} not {1.__class__!r}.'.format(PageSource, source))

        self._source = source
        self.current_page = 0
        if self.message is not None:
            await source._prepare_once()
            await self.show_page(0)

    def should_add_reactions(self) -> bool:
        return self.should_add_reactions_or_buttons()

    def should_add_reactions_or_buttons(self) -> bool:
        return self._source.is_paginating()

    async def _get_kwargs_from_page(self, page: List[Any]) -> SendKwargsType:
        value: PageFormatType = await nextcord.utils.maybe_coroutine(self._source.format_page, self, page)
        if isinstance(value, dict):
            return value
        elif isinstance(value, str):
            return {'content': value, 'embed': None}
        elif isinstance(value, nextcord.Embed):
            return {'embed': value, 'content': None}

    async def show_page(self, page_number: int):
        page = await self._source.get_page(page_number)
        self.current_page = page_number
        kwargs = await self._get_kwargs_from_page(page)
        await self.message.edit(**kwargs)

    async def send_initial_message(self, ctx: commands.Context, channel: nextcord.abc.Messageable) -> nextcord.Message:
        """|coro|

        The default implementation of :meth:`Menu.send_initial_message`
        for the interactive pagination session.

        This implementation shows the first page of the source.
        """
        page = await self._source.get_page(0)
        kwargs = await self._get_kwargs_from_page(page)
        if hasattr(self, '__nextcord_ui_view__'):
            kwargs['view'] = self
        return await channel.send(**kwargs)

    async def start(self, ctx: commands.Context, *, channel: Optional[nextcord.abc.Messageable] = None, wait: Optional[bool] = False):
        await self._source._prepare_once()
        await super().start(ctx, channel=channel, wait=wait)

    async def show_checked_page(self, page_number: int):
        max_pages = self._source.get_max_pages()
        try:
            # If it doesn't give maximum pages, it cannot be checked
            if max_pages is None or max_pages > page_number >= 0:
                await self.show_page(page_number)
        except IndexError:
            # An error happened that can be handled, so ignore it.
            pass

    async def show_current_page(self):
        if self._source.is_paginating():
            await self.show_page(self.current_page)

    def _skip_double_triangle_buttons(self) -> bool:
        max_pages = self._source.get_max_pages()
        if max_pages is None:
            return True
        return max_pages <= 2

    async def go_to_first_page(self, payload=None):
        """go to the first page"""
        await self.show_page(0)

    async def go_to_previous_page(self, payload=None):
        """go to the previous page"""
        await self.show_checked_page(self.current_page - 1)

    async def go_to_next_page(self, payload=None):
        """go to the next page"""
        await self.show_checked_page(self.current_page + 1)

    async def go_to_last_page(self, payload=None):
        """go to the last page"""
        # The call here is safe because it's guarded by skip_if
        await self.show_page(self._source.get_max_pages() - 1)

    async def stop_pages(self, payload=None):
        """stops the pagination session."""
        self.stop()


class MenuPages(MenuPagesBase):
    """A special type of Menu dedicated to pagination with reactions.

    Attributes
    ------------
    current_page: :class:`int`
        The current page that we are in. Zero-indexed
        between [0, :attr:`PageSource.max_pages`).
    """

    def __init__(self, source: PageSource, **kwargs):
        super().__init__(source, **kwargs)
        # skip adding buttons if inherit_buttons=False was passed to metaclass
        if not self.__inherit_buttons__:
            return
        # add pagination reaction buttons
        buttons = (
            Button(self.FIRST_PAGE, self.go_to_first_page,
                   position=First(0), skip_if=self._skip_double_triangle_buttons),
            Button(self.PREVIOUS_PAGE,
                   self.go_to_previous_page, position=First(1)),
            Button(self.NEXT_PAGE, self.go_to_next_page, position=Last(0)),
            Button(self.LAST_PAGE, self.go_to_last_page,
                   position=Last(1), skip_if=self._skip_double_triangle_buttons),
            Button(self.STOP, self.stop_pages, position=Last(2)),
        )
        for button in buttons:
            self.add_button(button)


class MenuPaginationButton(nextcord.ui.Button['MenuPaginationButton']):
    """
    A custom button for pagination that will be disabled when unavailable.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        emoji = kwargs.get("emoji", None)
        self._emoji = _cast_emoji(emoji) if emoji else None

    async def callback(self, interaction: nextcord.Interaction):
        """
        Callback for when this button is pressed
        """
        if self._emoji is None:
            return

        assert self.view is not None
        view: ButtonMenuPages = self.view

        # change the current page
        if str(self._emoji) == view.FIRST_PAGE:
            await view.go_to_first_page()
        elif str(self._emoji) == view.PREVIOUS_PAGE:
            await view.go_to_previous_page()
        elif str(self._emoji) == view.NEXT_PAGE:
            await view.go_to_next_page()
        elif str(self._emoji) == view.LAST_PAGE:
            await view.go_to_last_page()

        # disable buttons that are unavailable
        view._disable_unavailable_buttons()

        # disable all buttons if stop is pressed
        if str(self._emoji) == view.STOP:
            return view.stop()

        # update the view
        await interaction.response.edit_message(view=view)


class ButtonMenuPages(MenuPagesBase, ButtonMenu):
    """A special type of Menu dedicated to pagination with button components.

    Parameters
    -----------
    style: :class:`nextcord.ui.ButtonStyle`
        The button style to use for the pagination buttons.

    Attributes
    ------------
    current_page: :class:`int`
        The current page that we are in. Zero-indexed
        between [0, :attr:`PageSource.max_pages`).
    """

    def __init__(self, source: PageSource, style: nextcord.ButtonStyle = nextcord.ButtonStyle.secondary, **kwargs):
        self.__button_menu_pages__ = True
        # make button pagination disable buttons on stop by default unless it's overridden
        if "disable_buttons_after" not in kwargs:
            kwargs["disable_buttons_after"] = True
        super().__init__(source, **kwargs)
        # skip adding buttons if inherit_buttons=False was passed to metaclass
        if not self.__inherit_buttons__:
            return
        # add buttons to the view
        for emoji in (self.FIRST_PAGE, self.PREVIOUS_PAGE, self.NEXT_PAGE, self.LAST_PAGE, self.STOP):
            if emoji in {self.FIRST_PAGE, self.LAST_PAGE} and self._skip_double_triangle_buttons():
                continue
            self.add_item(MenuPaginationButton(emoji=emoji, style=style))
        self._disable_unavailable_buttons()

    def _disable_unavailable_buttons(self):
        """
        Disables buttons that are unavailable to be pressed.
        """
        children: List[MenuPaginationButton] = self.children
        max_pages = self._source.get_max_pages()
        for child in children:
            if isinstance(child, nextcord.ui.Button):
                if str(child.emoji) in (self.FIRST_PAGE, self.PREVIOUS_PAGE):
                    child.disabled = self.current_page == 0
                elif max_pages and str(child.emoji) in (self.LAST_PAGE, self.NEXT_PAGE):
                    child.disabled = self.current_page == max_pages - 1



"""An experiment that allows you to use a dict converter using **kwarg
notation.

.. note::
    This uses key, value pairs for construction

Specificy key and value types can be set using typing.Dict[key_type, value_type],
this defaults to Dict[str, str] the builtin dict type can also be used (that will default
to [str, str] aswell).

Example:
```py
import typing

@bot.command()
async def ban(ctx, **users_reasons_mapping: typing.Dict[nextcord.Member, str]):
    for (member, reason) in users_reasons_mapping['users_reasons_mapping'].items():  # this is necessary unfortunately
        await member.ban(reason=reason)
    await ctx.send(f'Banned {len(users_reasons_mapping["users_reasons_mapping"])} members')
```
"""

import inspect
from typing import Dict

import nextcord
from nextcord.ext.commands import Command, TooManyArguments, view as _view


class DictStringView(_view.StringView):
    def get_quoted_word(self):
        current = self.current
        if current is None:
            return None

        close_quote = _view._quotes.get(current)
        is_quoted = bool(close_quote)
        if is_quoted:
            result = []
            _escaped_quotes = (current, close_quote)
        else:
            result = [current]
            _escaped_quotes = _view._all_quotes

        while not self.eof:
            current = self.get()
            if not current:
                if is_quoted:
                    # unexpected EOF
                    raise _view.ExpectedClosingQuoteError(close_quote)
                return "".join(result)

            # currently we accept strings in the format of "hello world"
            # to embed a quote inside the string you must escape it: "a \"world\""
            if current == "\\":
                next_char = self.get()
                if not next_char:
                    # string ends with \ and no character after it
                    if is_quoted:
                        # if we're quoted then we're expecting a closing quote
                        raise _view.ExpectedClosingQuoteError(close_quote)
                    # if we aren't then we just let it through
                    return "".join(result)

                if next_char in _escaped_quotes:
                    # escaped quote
                    result.append(next_char)
                else:
                    # different escape character, ignore it
                    self.undo()
                    result.append(current)
                continue

            if not is_quoted and current in _view._all_quotes:
                # we aren't quoted
                try:
                    return self.get_quoted_word()
                except _view.UnexpectedQuoteError:
                    raise

            # closing quote
            if is_quoted and current == close_quote:
                next_char = self.get()
                # all this for that
                valid_eof = not next_char or next_char.isspace() or next_char == "="
                if not valid_eof:
                    raise _view.InvalidEndOfQuotedStringError(next_char)

                # we're quoted so it's okay
                return "".join(result)

            if current.isspace() and not is_quoted:
                # end of word found
                return "".join(result)

            result.append(current)


async def _parse_arguments(self, ctx):
    ctx.args = [ctx] if self.cog is None else [self.cog, ctx]
    ctx.kwargs = {}
    args = ctx.args
    kwargs = ctx.kwargs

    view = ctx.view
    iterator = iter(self.params.items())

    if self.cog is not None:
        # we have 'self' as the first parameter so just advance
        # the iterator and resume parsing
        try:
            next(iterator)
        except StopIteration:
            fmt = 'Callback for {0.name} command is missing "self" parameter.'
            raise nextcord.ClientException(fmt.format(self))

    # next we have the 'ctx' as the next parameter
    try:
        next(iterator)
    except StopIteration:
        fmt = 'Callback for {0.name} command is missing "ctx" parameter.'
        raise nextcord.ClientException(fmt.format(self))

    for name, param in iterator:
        if param.kind == param.POSITIONAL_OR_KEYWORD:
            transformed = await self.transform(ctx, param)
            args.append(transformed)
        elif param.kind == param.KEYWORD_ONLY:
            # kwarg only param denotes "consume rest" semantics
            if self.rest_is_raw:
                converter = self._get_converter(param)
                argument = view.read_rest()
                kwargs[name] = await self.do_conversion(ctx, converter, argument, param)
            else:
                kwargs[name] = await self.transform(ctx, param)
            break
        elif param.kind == param.VAR_POSITIONAL:
            while not view.eof:
                try:
                    transformed = await self.transform(ctx, param)
                    args.append(transformed)
                except RuntimeError:
                    break
        elif param.kind == param.VAR_KEYWORD:
            # we have received **kwargs
            annotation = param.annotation
            if annotation == param.empty or annotation is dict:
                annotation = Dict[str, str]  # default to {str: str}

            key_converter = annotation.__args__[0]
            value_converter = annotation.__args__[1]
            argument = view.read_rest()
            view = DictStringView(argument)
            kv_list = []

            while 1:
                kv = view.get_quoted_word()
                if kv is None:
                    break
                else:
                    kv_list.append(kv.strip())
            kv_pairs = []
            for current in kv_list:
                if current[0] == "=":
                    kv_pairs.remove([previous])
                    kv_pairs.append([previous, current[1:]])
                else:
                    kv_pairs.append(current.split("="))
                previous = current
            kwargs[name] = {
                await self.do_conversion(ctx, key_converter, key, param): await self.do_conversion(
                    ctx, value_converter, value, param
                )
                for (key, value) in kv_pairs
            }
            break

    if not self.ignore_extra:
        if not view.eof:
            raise TooManyArguments("Too many arguments passed to " + self.qualified_name)


Command._parse_arguments = _parse_arguments



"""An experiment that allows the use of returns in the command callback to reply to the user

Example:
```py
@bot.command()
async def foo(ctx):
    await ctx.send('hello')
    return 'world!'

'?foo' -->

'hello'
'world!'
"""

from nextcord.ext import commands


async def invoke(self, ctx):
    await self.prepare(ctx)

    ctx.invoked_subcommand = None
    injected = commands.core.hooked_wrapped_callback(self, ctx, self.callback)

    ret = await injected(*ctx.args, **ctx.kwargs)
    if ret is not None:
        await ctx.send(ret)


commands.Command.invoke = invoke



from nextcord.gateway import nextcordWebSocket


class BunsenBurner(nextcordWebSocket):
    @classmethod
    async def from_client(cls, client, **kwargs):
        client.analyst.ready_timer.start()
        return await super().from_client(client, **kwargs)

    async def send_as_json(self, data):
        self._dispatch('socket_send', data)
        await super().send_as_json(data)



"""An experiment that enables ``__int__`` on certain objects
to return the ``id`` attribute.
"""

import nextcord


_int = lambda self: self.id

nextcord.AppInfo.__int__ = _int
nextcord.Attachment.__int__ = _int
nextcord.AuditLogEntry.__int__ = _int
nextcord.emoji._EmojiTag.__int__ = _int
nextcord.mixins.Hashable.__int__ = _int
nextcord.Member.__int__ = _int
nextcord.Message.__int__ = _int
nextcord.Reaction.__int__ = _int
nextcord.Team.__int__ = _int
nextcord.Webhook.__int__ = _int



import inspect
import itertools
from typing import (Any, AsyncIterator, Callable, List, NamedTuple, Optional,
                    Sequence, TypeVar, Union)

from .constants import PageFormatType, SendKwargsType
from .menus import Menu

DataType = TypeVar("DataType")


class PageSource:
    """An interface representing a menu page's data source for the actual menu page.

    Subclasses must implement the backing resource along with the following methods:

    - :meth:`get_page`
    - :meth:`is_paginating`
    - :meth:`format_page`
    """
    async def _prepare_once(self):
        try:
            # Don't feel like formatting hasattr with
            # the proper mangling
            # read this as follows:
            # if hasattr(self, '__prepare')
            # except that it works as you expect
            self.__prepare
        except AttributeError:
            await self.prepare()
            self.__prepare = True

    async def prepare(self):
        """|coro|

        A coroutine that is called after initialisation
        but before anything else to do some asynchronous set up
        as well as the one provided in ``__init__``.

        By default this does nothing.

        This coroutine will only be called once.
        """
        pass

    def is_paginating(self) -> bool:
        """An abstract method that notifies the :class:`MenuPagesBase` whether or not
        to start paginating. This signals whether to add reactions or not.

        Subclasses must implement this.

        Returns
        --------
        :class:`bool`
            Whether to trigger pagination.
        """
        raise NotImplementedError

    def get_max_pages(self) -> Optional[int]:
        """An optional abstract method that retrieves the maximum number of pages
        this page source has. Useful for UX purposes.

        The default implementation returns ``None``.

        Returns
        --------
        Optional[:class:`int`]
            The maximum number of pages required to properly
            paginate the elements, if given.
        """
        return None

    async def get_page(self, page_number: int) -> Any:
        """|coro|

        An abstract method that retrieves an object representing the object to format.

        Subclasses must implement this.

        .. note::

            The page_number is zero-indexed between [0, :meth:`get_max_pages`),
            if there is a maximum number of pages.

        Parameters
        -----------
        page_number: :class:`int`
            The page number to access.

        Returns
        ---------
        Any
            The object represented by that page.
            This is passed into :meth:`format_page`.
        """
        raise NotImplementedError

    async def format_page(self, menu: Menu, page: Any) -> PageFormatType:
        """|maybecoro|

        An abstract method to format the page.

        This method must return one of the following types.

        If this method returns a ``str`` then it is interpreted as returning
        the ``content`` keyword argument in :meth:`nextcord.Message.edit`
        and :meth:`nextcord.abc.Messageable.send`.

        If this method returns a :class:`nextcord.Embed` then it is interpreted
        as returning the ``embed`` keyword argument in :meth:`nextcord.Message.edit`
        and :meth:`nextcord.abc.Messageable.send`.

        If this method returns a ``dict`` then it is interpreted as the
        keyword-arguments that are used in both :meth:`nextcord.Message.edit`
        and :meth:`nextcord.abc.Messageable.send`. The two of interest are
        ``embed`` and ``content``.

        Parameters
        ------------
        menu: :class:`Menu`
            The menu that wants to format this page.
        page: Any
            The page returned by :meth:`PageSource.get_page`.

        Returns
        ---------
        Union[:class:`str`, :class:`nextcord.Embed`, :class:`dict`]
            See above.
        """
        raise NotImplementedError


class ListPageSource(PageSource):
    """A data source for a sequence of items.

    This page source does not handle any sort of formatting, leaving it up
    to the user. To do so, implement the :meth:`format_page` method.

    Attributes
    ------------
    entries: Sequence[Any]
        The sequence of items to paginate.
    per_page: :class:`int`
        How many elements are in a page.
    """

    def __init__(self, entries: Sequence[DataType], *, per_page: int):
        self.entries = entries
        self.per_page = per_page

        pages, left_over = divmod(len(entries), per_page)
        if left_over:
            pages += 1

        self._max_pages = pages

    def is_paginating(self) -> bool:
        """:class:`bool`: Whether pagination is required."""
        return len(self.entries) > self.per_page

    def get_max_pages(self) -> int:
        """:class:`int`: The maximum number of pages required to paginate this sequence."""
        return self._max_pages

    async def get_page(self, page_number: int) -> Union[DataType, List[DataType]]:
        """Returns either a single element of the sequence or
        a slice of the sequence.

        If :attr:`per_page` is set to ``1`` then this returns a single
        element. Otherwise it returns at most :attr:`per_page` elements.

        Returns
        ---------
        Union[Any, List[Any]]
            The data returned.
        """
        if self.per_page == 1:
            return self.entries[page_number]
        else:
            base = page_number * self.per_page
            return self.entries[base:base + self.per_page]


KeyType = TypeVar("KeyType")

KeyFuncType = Callable[[DataType], KeyType]


class _GroupByEntry(NamedTuple):
    key: KeyFuncType
    items: DataType


class GroupByPageSource(ListPageSource):
    """A data source for grouped by sequence of items.

    This inherits from :class:`ListPageSource`.

    This page source does not handle any sort of formatting, leaving it up
    to the user. To do so, implement the :meth:`format_page` method.

    Parameters
    ------------
    entries: Sequence[Any]
        The sequence of items to paginate and group.
    key: Callable[[Any], Any]
        A key function to do the grouping with.
    sort: :class:`bool`
        Whether to sort the sequence before grouping it.
        The elements are sorted according to the ``key`` function passed.
    per_page: :class:`int`
        How many elements to have per page of the group.
    """

    def __init__(self, entries: Sequence[DataType], *, key: KeyFuncType, per_page: int, sort: int = True):
        self.__entries = entries if not sort else sorted(entries, key=key)
        nested: List[_GroupByEntry] = []
        self.nested_per_page = per_page
        for key_i, group_i in itertools.groupby(self.__entries, key=key):
            group_i = list(group_i)
            if not group_i:
                continue
            size = len(group_i)

            # Chunk the nested pages
            nested.extend(
                _GroupByEntry(key=key_i, items=group_i[i:i+per_page])
                for i in range(0, size, per_page)
            )

        super().__init__(nested, per_page=1)

    async def get_page(self, page_number: int) -> DataType:
        return self.entries[page_number]

    async def format_page(self, menu: Menu, entry: _GroupByEntry) -> SendKwargsType:
        """An abstract method to format the page.

        This works similar to the :meth:`ListPageSource.format_page` except
        the return type of the ``entry`` parameter is documented.

        Parameters
        ------------
        menu: :class:`Menu`
            The menu that wants to format this page.
        entry
            A namedtuple with ``(key, items)`` representing the key of the
            group by function and a sequence of paginated items within that
            group.

        Returns
        ---------
        :class:`dict`
            A dictionary representing keyword-arguments to pass to
            the message related calls.
        """
        raise NotImplementedError


def _aiter(obj, *, _isasync=inspect.iscoroutinefunction):
    cls = obj.__class__
    try:
        async_iter = cls.__aiter__
    except AttributeError:
        raise TypeError(
            '{0.__name__!r} object is not an async iterable'.format(cls))

    async_iter = async_iter(obj)
    if _isasync(async_iter):
        raise TypeError(
            '{0.__name__!r} object is not an async iterable'.format(cls))
    return async_iter


class AsyncIteratorPageSource(PageSource):
    """A data source for data backed by an asynchronous iterator.

    This page source does not handle any sort of formatting, leaving it up
    to the user. To do so, implement the :meth:`format_page` method.

    Parameters
    ------------
    iterator: AsyncIterator[Any]
        The asynchronous iterator to paginate.
    per_page: :class:`int`
        How many elements to have per page.
    """

    def __init__(self, iterator: AsyncIterator[DataType], *, per_page: int):
        self.iterator = _aiter(iterator)
        self.per_page = per_page
        self._exhausted = False
        self._cache: List[DataType] = []

    async def _iterate(self, n: int):
        it = self.iterator
        cache = self._cache
        for _ in range(0, n):
            try:
                elem = await it.__anext__()
            except StopAsyncIteration:
                self._exhausted = True
                break
            else:
                cache.append(elem)

    async def prepare(self, *, _aiter=_aiter):
        # Iterate until we have at least a bit more single page
        await self._iterate(self.per_page + 1)

    def is_paginating(self) -> bool:
        """:class:`bool`: Whether pagination is required."""
        return len(self._cache) > self.per_page

    async def _get_single_page(self, page_number: int) -> DataType:
        if page_number < 0:
            raise IndexError('Negative page number.')

        if not self._exhausted and len(self._cache) <= page_number:
            await self._iterate((page_number + 1) - len(self._cache))
        return self._cache[page_number]

    async def _get_page_range(self, page_number: int) -> List[DataType]:
        if page_number < 0:
            raise IndexError('Negative page number.')

        base = page_number * self.per_page
        max_base = base + self.per_page
        if not self._exhausted and len(self._cache) <= max_base:
            await self._iterate((max_base + 1) - len(self._cache))

        entries = self._cache[base:max_base]
        if not entries and max_base > len(self._cache):
            raise IndexError('Went too far')
        return entries

    async def get_page(self, page_number: int) -> Union[DataType, List[DataType]]:
        """Returns either a single element of the sequence or
        a slice of the sequence.

        If :attr:`per_page` is set to ``1`` then this returns a single
        element. Otherwise it returns at most :attr:`per_page` elements.

        Returns
        ---------
        Union[Any, List[Any]]
            The data returned.
        """
        if self.per_page == 1:
            return await self._get_single_page(page_number)
        else:
            return await self._get_page_range(page_number)



"""An experiment that allows you to use the `&` and `|` operators
on checks, allowing for easier control over check conditions.

This should not break pre-existing checks.

Examples:
```py
@guild_only() | dm_only()
@b.command()
async def there(ctx):
    await ctx.send('is literally no reason to use both the guild_only and dm_only checks on the same command.')
```

```py
@dm_only() | is_owner()
@bot.command()
async def hello(ctx):
    await ctx.send('world!')
```

```py
@guild_only() & ((has_role('Capitalist') & bot_has_role('Leader of the Communist Revolution') ) | !is_owner())
@bot.command()
async def no(ctx):
    await ctx.send('stop this')
```
"""

import inspect

from nextcord.ext import commands
from nextcord.ext.alternatives._common import py_allow


py_allow(3, 9, 0)


class CheckDecorator:
    def __init__(self, predicate):
        self.predicate = predicate
        self.check = Only(Check(predicate))

    def __call__(self, func):
        if isinstance(func, commands.Command):
            func.checks.append(self.check)
        else:
            if not hasattr(func, "__commands_checks__"):
                func.__commands_checks__ = []

            func.__commands_checks__.append(self.check)

        return func

    def __repr__(self):
        return f"CheckDecorator<check={self.check!r}>"

    def __invert__(self):
        ~self.check.first
        return self.check

    def __or__(self, other):
        self.check.first = Either(
            self.check.first, other.check.first if isinstance(other, CheckDecorator) else other
        )
        return self.check

    def __and__(self, other):
        self.check.first = Both(
            self.check.first, other.check.first if isinstance(other, CheckDecorator) else other
        )
        return self.check


class Check:
    def __init__(self, predicate):
        self.predicate = predicate
        self.inverted = False

    def __repr__(self):
        return f"Check(predicate={self.predicate!r}, inverted={self.inverted})"

    async def __call__(self, *args, **kwargs):
        r = self.predicate(*args, **kwargs)
        if isinstance(r, bool):
            r = r
        else:
            r = await r

        if self.inverted:
            r = not r

        return r

    def __invert__(self):
        self.inverted = not self.inverted
        return self

    def __or__(self, other):
        return Either(self.predicate, other if isinstance(other, CheckOp) else other.predicate)

    def __and__(self, other):
        return Both(self.predicate, other if isinstance(other, CheckOp) else other.predicate)


commands.core.check = CheckDecorator
commands.check = CheckDecorator


class Only:
    def __init__(self, first: Check):
        self.first = first
        self.inverted = False

    def _call(self, *args, **kwargs):
        return self.first(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        if isinstance(args[0], commands.Command):
            args[0].checks.append(self)
            return args[0]
        else:
            return self._call(*args, **kwargs)

    def __repr__(self):
        return f"Only(first={self.first!r})"

    def __invert__(self):
        self.inverted = not self.inverted
        return self

    def __or__(self, other):
        return Either(self, other)

    def __and__(self, other):
        return Both(self, other)


class CheckOp:
    def __init__(self, first: Check, second: Check):
        self.first = first
        self.second = second
        self.inverted = False
        self.check = self

    def __repr__(self):
        return f"{self.__class__.__name__}(first={self.first!r}, second={self.second!r}, inverted={self.inverted})"

    async def _try_single(self, callback, *args, **kwargs):
        r = await callback(*args, **kwargs)

        return not r if self.inverted else r

    async def _try_call(self, *args, **kwargs):
        return await self._try_single(self.first, *args, **kwargs), await self._try_single(
            self.second, *args, **kwargs
        )

    def __invert__(self):
        self.inverted = not self.inverted
        return self

    def __or__(self, other):
        return Either(self, other)

    def __and__(self, other):
        return Both(self, other)

    async def _call(self, *args, **kwargs):
        ...

    def __call__(self, *args, **kwargs):
        if isinstance(args[0], commands.Command):
            args[0].checks.append(self)
            return args[0]
        else:
            return self._call(*args, **kwargs)


class Both(CheckOp):
    async def _call(self, *args, **kwargs):
        fs, ss = await self._try_call(*args, **kwargs)
        return fs and ss


class Either(CheckOp):
    async def _call(self, *args, **kwargs):
        try:
            if await self._try_single(self.first, *args, **kwargs):
                return True
        except:
            pass

        try:
            if await self._try_single(self.second, *args, **kwargs):
                return True
        except:
            pass

        return False



"""**STANDALONE**: An experiment that allows the use of classes and
functions as a way to represent a group of commands.

Example:
```py
@bot.group(cls=ClassGroup)
class A(Config, invoke_without_command=True): # A group command
    _CONFIG = Config(invoke_without_command=True, description='no')

    async def __call__(ctx):
        await ctx.send('test')
    
    async def fmt(ctx): # A command
        await ctx.send('toot')
    
    class B: # A group command
        #_CONFIG = Config(description='yes')

        async def __call__(ctx):
            await ctx.send('no')

        async def oops(ctx): # A command
            await ctx.send('yert')
```

```
!A -> 'test'
!A fmt -> 'toot'
!A B -> 'no'
!A B oops -> 'no', 'yert'
```
"""

import inspect

from nextcord.ext import commands


class ClassGroup(commands.Group):
    def __init__(self, cls, *, name=None, parent=None):
        kwargs = {"name": name or cls.__name__, "parent": parent}
        func = cls.__call__

        try:
            cls._CONFIG
        except AttributeError:
            pass
        else:
            kwargs.update(cls._CONFIG.to_dict())

        super().__init__(func, **kwargs)

        for f in dir(cls):
            if f.startswith("_"):
                continue

            attr = getattr(cls, f)

            if inspect.isclass(attr):
                self.add_command(ClassGroup(attr, parent=self))
            elif inspect.iscoroutinefunction(attr):
                self.add_command(commands.Command(attr))


class Config:
    def __init__(
        self,
        *,
        invoke_without_command: bool = False,
        case_insensitive: bool = False,
        help: str = "",
        brief: str = "",
        usage: str = "",
        aliases: list = [],
        checks: list = [],
        description: str = "",
        hidden: bool = False,
    ):
        self.invoke_without_command = invoke_without_command
        self.case_insensitive = case_insensitive
        self.help = help
        self.brief = brief
        self.usage = usage
        self.aliases = aliases
        self.checks = checks
        self.description = description
        self.hidden = hidden

    def to_dict(self):
        d = {}
        for attr in dir(self):
            if not attr.startswith("_"):
                d[attr] = getattr(self, attr)

        return d



"""An experiment that enables auto removing reactions added to menus.

"""

import nextcord
from nextcord.ext import menus


_old_update = menus.Menu.update


async def update(self: menus.Menu, payload: nextcord.RawReactionActionEvent):
    await _old_update(self, payload)

    if payload.event_type != "REACTION_ADD":
        return

    permissions = self.ctx.channel.permissions_for(self.ctx.me)
    if not (permissions.manage_messages or permissions.administrator):
        return

    await self.message.remove_reaction(payload.emoji, nextcord.Object(id=payload.user_id))


update.__doc__ = _old_update.__doc__
menus.Menu.update = update



from typing import Pattern
from .constants import EmojiType
import re

import nextcord


class Position:
    __slots__ = ('number', 'bucket')

    def __init__(self, number: int, *, bucket: int = 1):
        self.bucket = bucket
        self.number = number

    def __lt__(self, other):
        if not isinstance(other, Position) or not isinstance(self, Position):
            return NotImplemented

        return (self.bucket, self.number) < (other.bucket, other.number)

    def __eq__(self, other):
        return isinstance(other, Position) and other.bucket == self.bucket and other.number == self.number

    def __le__(self, other):
        r = Position.__lt__(other, self)
        if r is NotImplemented:
            return NotImplemented
        return not r

    def __gt__(self, other):
        return Position.__lt__(other, self)

    def __ge__(self, other):
        r = Position.__lt__(self, other)
        if r is NotImplemented:
            return NotImplemented
        return not r

    def __repr__(self):
        return '<{0.__class__.__name__}: {0.number}>'.format(self)


class Last(Position):
    __slots__ = ()

    def __init__(self, number=0):
        super().__init__(number, bucket=2)


class First(Position):
    __slots__ = ()

    def __init__(self, number=0):
        super().__init__(number, bucket=0)


_custom_emoji = re.compile(
    r'<?(?P<animated>a)?:?(?P<name>[A-Za-z0-9\_]+):(?P<id>[0-9]{13,20})>?')


def _cast_emoji(obj: EmojiType, *, _custom_emoji: Pattern[str]=_custom_emoji):
    if isinstance(obj, nextcord.PartialEmoji):
        return obj

    obj = str(obj)
    match = _custom_emoji.match(obj)
    if match is not None:
        groups = match.groupdict()
        animated = bool(groups['animated'])
        emoji_id = int(groups['id'])
        name = groups['name']
        return nextcord.PartialEmoji(name=name, animated=animated, id=emoji_id)
    return nextcord.PartialEmoji(name=obj, id=None, animated=False)



import typing
from nextcord.ext import commands
import copy

class CustomConveterMeta(type):
    def __getitem__(cls, value: typing.Any) -> 'CustomConverter':
        klass = copy.deepcopy(cls)
        klass._value = value
        return klass

class CustomConverter(commands.Converter, metaclass=CustomConveterMeta):
    _value: typing.Any



"""An experiment that enables ``Guild.jump_url`` and ``abc.Messagable.jump_url``.

Example:

```py
>>> message.channel.jump_url
https://nextcord.com/channels/364412422540361729/708518465820033105
```
"""

import nextcord


@property
def guild_jump_url(self):
    """:class:`str`: Returns a URL that allows the client to jump to this guild."""

    return "https://nextcord.com/channels/{0.id}".format(self)


nextcord.Guild.jump_url = guild_jump_url


@property
def messageable_jump_url(self):
    """:class:`str`: Returns a URL that allows the client to jump to this channel."""

    if isinstance(self, nextcord.abc.User):
        if self.dm_channel is None:
            raise AttributeError("Could not find DM channel for user '{0}'".format(self))

        channel_id = self.dm_channel.id
    else:
        channel_id = self.channel.id if hasattr(self, "channel") else self.id

    guild_id = self.guild.id if isinstance(self, nextcord.abc.GuildChannel) else "@me"

    return "https://nextcord.com/channels/{0}/{1}".format(guild_id, channel_id)


nextcord.abc.Messageable.jump_url = messageable_jump_url



"""An experiment that allows you to shave a role off of all members. Optionally,
you can pass Member objects (or IDs) to spare those members from having the role shaved.
"""

from nextcord import Role
from nextcord.abc import Snowflake
from typing import Union, Optional


async def _shave(
    self,
    *except_members: Union[int, Snowflake],
    reason: Optional[str] = None,
    atomic: bool = True,
):
    r"""|coro|

    Shaves :class:`Role`\s from all members with the role, unless explicitly
    added to except_members.

    You must have the :attr:`~Permissions.manage_roles` permission to
    use this.

    Parameters
    -----------
    \*except_members: :class:`abc.Snowflake`
        An argument list of :class:`abc.Snowflake` representing a :class:`Member`
        to not shave the role from.
    reason: Optional[:class:`str`]
        The reason for removing these roles. Shows up on the audit log.
    atomic: :class:`bool`
        Whether to atomically remove roles. This will ensure that multiple
        operations will always be applied regardless of the current
        state of the cache.

    Raises
    -------
    Forbidden
        You do not have permissions to remove these roles.
    HTTPException
        Removing the roles failed.
    """

    except_members_ids = {int(m) for m in except_members}

    for member in self.members:
        if member.id in except_members_ids:
            continue

        await member.remove_roles(self, reason=reason, atomic=atomic)


Role.shave = _shave



"""An experiment that allows you to send help without requiring ``Context``,
or inside of ``on_message``.

Example:
```py
@bot.event
async def on_message(message):
    if message.content == message.guild.me.mention:
        await bot.send_help(message) # sends the entire help command.
```
"""

from nextcord.ext import commands


def send_help(self, message, *args, **kwargs):
    ctx = kwargs.get("cls", commands.Context)(prefix=self.user.mention, bot=self, message=message)
    return ctx.send_help(*args)


commands.bot.BotBase.send_help = send_help



from typing import Literal, get_args, get_origin
from nextcord.ext.commands import Command
from nextcord.ext.commands.errors import ConversionError, BadArgument

_old_actual_conversion = Command._actual_conversion


async def _actual_conversion(self, ctx, converter, argument, param):
    origin = get_origin(converter)

    if origin is Literal:
        items = get_args(converter)

        if all(i for i in items if isinstance(i, str)):
            if argument in items:
                return argument

            raise BadArgument(f"Expected literal: one of {list(map(repr, items))}")
        elif all(i for i in items if not isinstance(i, str)):
            ret = await _old_actual_conversion(self, ctx, type(items[0]), argument, param)
            return ret in items
        else:
            raise ConversionError("Literal contains multiple conflicting types.")

    return await _old_actual_conversion(self, ctx, converter, argument, param)


Command._actual_conversion = _actual_conversion



import inspect
import copy
import typing

from nextcord.ext.commands import GroupMixin, Command, Group

from .model import ConverterDict

__all__ = (
    'ConvertersCommand',
    'ConvertersGroupMixin',
    'ConvertersGroup',
    'command',
    'group',
)

class ConvertersCommand(Command):
    converters: ConverterDict

    async def _actual_conversion(self, ctx, converter, argument, param):
        converter = self.converters.get(converter, converter)
        return await super()._actual_conversion(ctx, converter, argument, param)

T = ConvertersCommand

class ConvertersGroupMixin(GroupMixin):
    converters = ConverterDict()

    # redefine GroupMixin#command and GroupMixin#group to work with redefined decorators
    def command(self, *args, **kwargs) -> typing.Callable[..., T]:
        def decorator(func: typing.Callable[..., typing.Any]) -> T:
            kwargs.setdefault('parent', self)
            result = command(*args, **kwargs)(func)
            result.converters = copy.copy(self.converters)
            self.add_command(result)
            return result
        
        return decorator
    
    def group(self, *args, **kwargs) -> typing.Callable[..., T]:
        def decorator(func: typing.Callable[..., typing.Any]) -> T:
            kwargs.setdefault('parent', self)
            result = group(*args, **kwargs)(func)
            result.converters.update(copy.copy(self.converters))
            self.add_command(result)
            return result

        return decorator


class ConvertersGroup(ConvertersGroupMixin, ConvertersCommand):
    ...

def command(name: typing.Optional[str] = None, cls: typing.Optional[typing.Type[T]] = None, **attrs: typing.Any) -> typing.Callable[..., T]:
    if cls is None:
        cls = ConvertersCommand

    def decorator(func: typing.Callable[..., typing.Any]) -> T:
        if isinstance(func, ConvertersCommand):
            raise TypeError('Callback is already a command.')
        return cls(func, name=name, **attrs)

    return decorator

def group(name: typing.Optional[str] = None, **attrs: typing.Any) -> typing.Callable[..., typing.Union[ConvertersGroup, T]]:
    """A decorator that transforms a function into a :class:`.Group`.

    This is similar to the :func:`.command` decorator but the ``cls``
    parameter is set to :class:`Group` by default.

    .. versionchanged:: 1.1
        The ``cls`` parameter can now be passed.
    """

    attrs.setdefault('cls', ConvertersGroup)
    return command(name=name, **attrs)



"""
    Copyright 2020 Ext-Creators
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
        http://www.apache.org/licenses/LICENSE-2.0
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from .bot import RxBot
from .utils import _

from collections import namedtuple


__version__ = "0.1.0"

_VersionInfo = namedtuple("_VersionInfo", "major minor micro releaselevel serial")
version_info = _VersionInfo(major=0, minor=1, micro=0, releaselevel="stable", serial=0)



from nextcord.abc import Messageable
from nextcord.message import Message
import nextcord
from nextcord.ext import commands


def wait_for(self, event, *, check=None, timeout=None):
    actual_wait_for = self._state.dispatch.__self__.wait_for

    if check is None:

        def check(*args):
            return True

    def actual_check(*args):
        for arg in args:
            if isinstance(arg, (nextcord.Message, commands.Context)):
                if arg.channel.id == self.id:
                    return check(*args)
            elif isinstance(arg, nextcord.abc.Messageable):
                if arg.id == self.id:
                    return check(*args)

    return actual_wait_for(event, check=actual_check, timeout=timeout)


nextcord.abc.Messageable.wait_for = wait_for



import typing

from nextcord.flags import fill_with_flags, flag_value, BaseFlags

__all__ = (
    'EventFlags',
)

class _auto_value:
    _CURR_VALUE = 0

    def __rlshift__(self, val) -> int:
        ret = val << self._CURR_VALUE
        self._CURR_VALUE += 1
        return ret

    def reset(self) -> None:
        self._CURR_VALUE = 0

_GLOBAL_AUTO_VALUE = _auto_value()


def auto(doc=None, *, last=False) -> flag_value:
    _bit = 1 << _GLOBAL_AUTO_VALUE
    if last:
        _GLOBAL_AUTO_VALUE.reset()

    def pred(self) -> int:
        return _bit
    
    pred.__doc__ = doc
    return flag_value(pred)

class propmethod:
    def __init__(self, func):
        self._func = func
        self._flag = None
    
    def __get__(self, instance, owner):
        if instance is None:
            def pred():
                cls = owner
                _self = cls.__new__(cls)
                _self.value = self._func(cls)
                return _self
            return pred

        self._flag = flag_value(lambda n: self._func(instance))
        return self._flag.__get__(instance, owner)
        
    def __set__(self, instance, value):
        if instance is not None:
            if self._flag is None:
                self._flag = flag_value(lambda n: self._func(instance))
            self._flag.__set__(instance, value)
        else:
            return


class BaseScienceFlags(BaseFlags):
    @propmethod
    def all(cls):
        bits = max(cls.VALID_FLAGS.values()).bit_length()
        return (1 << bits) - 1
    
    @propmethod
    def none(cls):
        return cls.DEFAULT_VALUE


@fill_with_flags()
class EventFlags(BaseScienceFlags):
    @propmethod
    def guilds(cls):
        return 127

    guild_join = auto('The on_guild_join event')
    guild_remove = auto('The on_guild_remove event')
    guild_available = auto('The on_guild_available event')
    guild_unavailable = auto('The on_guild_unavailable event')
    guild_channel_update = auto('The on_guild_channel_update event')
    guild_channel_create = auto('The on_guild_channel_delete event')
    guild_channel_delete = auto('The on_guild_channel_pins_update event')

    @propmethod
    def members(cls):
        return 896
    
    member_join = auto('The on_member_join event')
    member_remove = auto('The on_member_remove event')
    member_update = auto('The on_member_update event')

    @propmethod
    def bans(cls):
        return 3072
    
    member_ban = auto('The on_member_ban event')
    member_unban = auto('The on_member_unban event')

    @propmethod
    def emojis(cls):
        return 4096
    
    guild_emojis_update = auto('The on_guild_emojis_update event')

    @propmethod
    def integrations(cls):
        return 8192

    guild_integrations_update = auto('The on_guild_integrations_update event')

    @propmethod
    def webhooks(cls):
        return 16384
    
    webhooks_update = auto('The on_webhooks_update event')

    @propmethod
    def invites(cls):
        return 98304
    
    invite_create = auto('The on_invite_create event')
    invite_delete = auto('The on_invite_delete event')

    @propmethod
    def voice_states(cls):
        return 131072
    
    voice_state_update = auto('The on_voice_state_update event')

    @propmethod
    def presences(cls):
        return 262144
    
    presence_update = auto('The on_member_update event')

    @propmethod
    def messages(cls):
        return 7864320
    
    message = auto('The on_message event')
    message_update = auto('The on_message_update event')
    message_delete = auto('The on_message_delete event')
    private_channel_create = auto('The on_private_channel_create event')

    @propmethod
    def reactions(cls):
        return 58720256
    
    reaction_add = auto('The on_reaction_add event')
    reaction_remove = auto('The on_reaction_remove event')
    reaction_clear = auto('The on_reaction_clear event')
    
    typing = auto('The on_typing event', last=True)


@fill_with_flags()
class OpFlags(BaseScienceFlags):
    DISPATCH = auto('An event sent to the Client (ie. READY).')
    HEARTBEAT = auto('To keep the connection alive.')
    IDENTIFY = auto('When a new session is started.')
    PRESENCE = auto('When you update your presence.')
    VOICE_STATE = auto('When a new connection is started to a voice guild.')
    VOICE_PING = auto('Checks the ping time to a voice guild.')
    RESUME = auto('When an existing connection is resumed.')
    RECONNECT = auto('When the client is told to reconnect.')
    REQUEST_MEMBERS = auto('When you ask for a full member list of a guild.')
    INVALIDATE_SESSION = auto('When the client is told to invalidate the session and re-IDENTIFY.')
    HELLO = auto('When the client is told the heartbeat interval.')
    HEARTBEAT_ACK = auto('When the heartbeat received is confirmed.')
    GUILD_SYNC = auto('When a guild sync is requested.', last=True)



"""
    Copyright 2021 Ext-Creators

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

"""An experiment that allows `x in y` syntax for various nextcord objects.

Example:
```py
channel in guild

member in channel

member in role
```
"""

import nextcord
from nextcord.channel import CategoryChannel, DMChannel, TextChannel

if nextcord.version_info < (1, 7, 0):
    from nextcord.channel import VoiceChannel as VocalGuildChannel
else:
    from nextcord.channel import VocalGuildChannel

from nextcord.guild import Guild
from nextcord.member import Member
from nextcord.message import Message
from nextcord.role import Role
from nextcord.user import User, BaseUser


def _Guild__contains__(self, item):
    if hasattr(item, "guild"):
        return item.guild == self

    if isinstance(item, BaseUser):
        return item.id in self._members

    return False


Guild.__contains__ = _Guild__contains__


def _Role__contains__(self, item):
    if isinstance(item, User):
        item = self.guild._members.get(item.id)

    if isinstance(item, Member):
        return item._roles.has(self.id)

    return False


Role.__contains__ = _Role__contains__


def _TextChannel__contains__(self, item):
    if hasattr(item, "channel"):
        return item.channel == self

    if isinstance(item, User):
        item = self.guild._members.get(item.id)

    if isinstance(item, Member):
        return self.permissions_for(item).read_messages

    return False


TextChannel.__contains__ = _TextChannel__contains__


def _VocalGuildChannel__contains__(self, item):
    if isinstance(item, BaseUser) and item.id in self.voice_states:
        return True

    return False


VocalGuildChannel.__contains__ = _VocalGuildChannel__contains__


def _CategoryChannel__contains__(self, item):
    if hasattr(item, "category"):
        return item.category == self

    return False


CategoryChannel.__contains__ = _CategoryChannel__contains__


def _DMChannel__contains__(self, item):
    if hasattr(item, "channel"):
        return item.channel == self

    if isinstance(item, BaseUser):
        return item in (self.me, self.recipient)

    return False


DMChannel.__contains__ = _DMChannel__contains__



"""
    Copyright (c) 2020 Ext-Creators

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
from .manager import GetchManager

class GetchBuilder:
    def __init__(self, *, bot):
        self.http = GetchManager(bot=bot)
        self.ids = {
            'guild': None,
            'channel': None,
            'message': None,
        }

    async def getch(self):
        guild = None
        channel = None
        message = None

        if (guild_id := self.ids['guild']):
            guild = self.http.get_guild(guild_id) or await self.http.fetch_guild(guild_id)

        if (channel_id := self.ids['channel']):
            channel = self.http.get_channel(channel_id) or await self.http.fetch_channel(channel_id, guild=guild)
        
        if (message_id := self.ids['message']):
            message = self.http.get_message(message_id) or await self.http.fetch_message(message_id, channel=channel)

        # return whatever object in order of depth.
        return message or channel or guild

    def uild(self, id: int) -> GetchBuilder:
        self.ids['guild'] = id
        return self

    def annel(self, id: int) -> GetchBuilder:
        self.ids['channel'] = id
        return self

    def essage(self, id: int) -> GetchBuilder:
        self.ids['message'] = id
        return self



"""An experiment that allows you to handle all exceptions of
group commands in the error handler of the root parent.

Example:
```py
@bot.group()
async def profile(ctx):
    pass

@profile.error
async def profile_error(ctx, error):
    if isinstance(error.original, Exception):
        await ctx.send("You have not created your profile yet. Use ;profile create command now to create your profile.")

@profile.command()
async def name(ctx):
    raise Exception

@profile.command()
async def color(ctx):
    raise Exception
```
"""

from nextcord.ext import commands


async def dispatch_error(self, ctx, error):
    ctx.command_failed = True
    cog = self.cog

    try:
        coro = self.on_error
    except AttributeError:
        pass
    else:
        injected = commands.core.wrap_callback(coro)

        if cog is not None:
            await injected(cog, ctx, error)
        else:
            await injected(ctx, error)

    try:
        coro = self.root_parent.on_error
    except AttributeError:
        pass
    else:
        injected = commands.core.wrap_callback(coro)

        if cog is not None:
            await injected(cog, ctx, error)
        else:
            await injected(ctx, error)

    try:
        if cog is not None:
            local = commands.Cog._get_overridden_method(cog.cog_command_error)
            if local is not None:
                wrapped = commands.core.wrap_callback(local)
                await wrapped(ctx, error)
    finally:
        ctx.bot.dispatch("command_error", ctx, error)


commands.Command.dispatch_error = dispatch_error



import asyncio
import datetime

from functools import wraps
from typing import Callable, Optional

import nextcord


SLEEP_FOR = 2.5


async def fetch_recent_audit_log_entry(client: nextcord.Client, guild: nextcord.Guild, *, target: nextcord.User = None,
                                       action: nextcord.AuditLogAction = None, retry: int = 0) -> Optional[nextcord.AuditLogEntry]:
    """|coro|

    Attempts to retrieve an recently created :class:`~nextcord.AuditLogEntry` which meets the specified requirements.

    Parameters
    ----------
    client: :class:`~nextcord.Client`
        The nextcord client to make the api calls with.
    guild: :class:`~nextcord.Guild`
        The guild to retrieve the audit log entry from.
    target: Optional[:class:`~nextcord.User`]
        The target to filter with.
    action: Optional[:class:`~nextcord.AuditLogAction`]
        The action to filter with.
    retry: Optional[:class:`int`]
        The number of times fetching an entry should be retried.
        Defaults to 0.

    Raises
    ------
    Forbidden
        You do not have access to the guild's audit log.
    HTTPException
        Fetching the member failed.

    Returns
    --------
    Optional[:class:`~nextcord.AuditLogEntry`]
        The relevant audit log entry if found.
    """
    async for entry in guild.audit_logs(limit=1, action=action):

        delta = datetime.datetime.utcnow() - entry.created_at
        if delta < datetime.timedelta(seconds=10):
            if target is not None and entry.target != target:
                continue

            return entry

    if retry > 0:
        await asyncio.sleep(SLEEP_FOR)
        return await fetch_recent_audit_log_entry(client, guild, target=target, action=action, retry=retry - 1)

    return None


def listens_for(*events: str) -> Callable:
    """Helper decorator which defines which events an event check is listening for.

    Parameters
    ----------
    *events: :class:`str`
        The names of the events to listen for.
    """

    def decorator(func: Callable) -> Callable:

        @wraps(func)
        async def wrapper(client, event, *args, **kwargs):
            if event in events:
                await func(client, *args, **kwargs)

        return wrapper

    return decorator



"""An experiment that allows for conversion of ``Guild``
arguments for commands.

Example:
```py
@commands.command()
async def test(ctx, server: Guild):
    await ctx.send(f"You selected **{server.name}**")
```
"""

from nextcord.ext.commands import BadArgument, converter, Context
from nextcord import Guild, utils

from ._common import _ALL

# Basic Guild Converter


class _GuildConverter(converter.IDConverter):
    async def convert(self, ctx: Context, argument: str):
        bot = ctx.bot

        match = self._get_id_match(argument)
        result = None

        if match is None:
            result = utils.get(bot.guilds, name=argument)
        else:
            guild_id = int(match.group(1))
            result = ctx.bot.get_guild(guild_id)

        if result is None:
            raise BadArgument('Guild "{}" not found'.format(argument))
        return result


converter.GuildConverter = _GuildConverter
_ALL[Guild] = _GuildConverter



import asyncio

__all__ = [
    '_'
]

def _(coro):
    """This function runs ensures the future of a coroutine (ie from a lambda).
    """
    def run(*args):
        return asyncio.ensure_future(coro(*args), loop=asyncio.get_event_loop())
    return run



import asyncio

import nextcord

from ._events import _ALL


class EventsMixin(nextcord.Client):

    async def on__event(self, event, *args, **kwargs):
        for event_name, event_check in _ALL.items():
            asyncio.ensure_future(event_check(self, event, *args, **kwargs))

    def dispatch(self, event, *args, **kwargs):
        super().dispatch(event, *args, **kwargs)  # type: ignore
        super().dispatch('_event', event, *args, **kwargs)



import collections

from .client import Client
from .server import Server
from .errors import *


_VersionInfo = collections.namedtuple("_VersionInfo", "major minor micro release serial")

version = "1.0.0"
version_info = _VersionInfo(1, 0, 0, "final", 0)



import re
import setuptools


classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Framework :: AsyncIO",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Communications",
    "Topic :: Documentation",
    "Topic :: Documentation :: Sphinx",
    "Topic :: Internet",
    "Topic :: Software Development",
    "Topic :: Software Development :: Libraries",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

"""with open("requirements.txt") as stream:
    install_requires = stream.read().splitlines() """

packages = [
    "nextcord.ext.ipc",
    "nextcord.ext.alternatives",
    "nextcord.ext.converters",
    "nextcord.ext.events",
    "nextcord.ext.getch",
    "nextcord.ext.menus",
    "nextcord.ext.rx",
    "nextcord.ext.science",
]

project_urls = {
    "Issue Tracker": "https://github.com/VincentRPS/nextcord-ext-all/issues",
    "Source": "https://github.com/VincentRPS/nextcord-ext-all",
}

_version_regex = r"^version = ('|\")((?:[0-9]+\.)*[0-9]+(?:\.?([a-z]+)(?:\.?[0-9])?)?)\1$"

with open("nextcord/ext/ipc/__init__.py") as stream:
    match = re.search(_version_regex, stream.read(), re.MULTILINE)

version = match.group(2)

if match.group(3) is not None:
    try:
        import subprocess

        process = subprocess.Popen(["git", "rev-list", "--count", "HEAD"], stdout=subprocess.PIPE)
        out, _ = process.communicate()
        if out:
            version += out.decode("utf-8").strip()

        process = subprocess.Popen(["git", "rev-parse", "--short", "HEAD"], stdout=subprocess.PIPE)
        out, _ = process.communicate()
        if out:
            version += "+g" + out.decode("utf-8").strip()
    except (Exception) as e:
        pass


setuptools.setup(
    author="Nextcord, Ext-Creators, VincentRPS",
    classifiers=classifiers,
    description="A Nextcord extension for all ext additions",
    license="Apache Software License",
    name="nextcord-ext-all",
    packages=packages,
    project_urls=project_urls,
    python_requires=">=3.6.0",
    url="https://github.com/VincentRPS/nextcord-ext-all",
    version=version,
)


import logging
import asyncio

from nextcord import Client
from nextcord.gateway import ResumeWebSocket

from .config import Configuration

logger = logging.getLogger(__name__)


class Scientist(Client):
    def __init__(self, *args, **kwargs):
        self.config: Configuration = kwargs.pop('config')
        logger.debug('Initiated Scientist with %s.', self.config)

        self.analyst = self.config.anal_cls(self.config.recorder, self.config.event_flags, self.config.op_flags)
        super().__init__(*args, **kwargs)
        self.http = self.config.httpclient_cls.from_http_client(self.http, self.config.session_cls, self.analyst)
    
    def dispatch(self, event, *args, **kwargs) -> None:
        self._schedule_event(self.analyst.log, 'science_logging', event, *args, **kwargs)
        super().dispatch(event, *args, **kwargs)
    
    async def start(self, *args, **kwargs) -> None:
        await self.config.recorder.start()
        next_id = await self.config.recorder.last_events_id() + 1
        logger.debug("The next ID for the 'events' table is %s", next_id)
        await super().start(*args, **kwargs)
    
    async def close(self) -> None:
        await super().close()
        if self.config.recorder.started:
            curr_id = await self.config.recorder.last_events_id()
            logger.debug("The final ID for the 'events' table is %s", curr_id)
            await self.config.recorder.end()
    
    async def _connect(self) -> None:
        cls = self.config.gw_cls
        coro = cls.from_client(self, shard_id=self.shard_id)
        self.ws: cls = await asyncio.wait_for(coro, timeout=180.0)
        while True:
            try:
                await self.ws.poll_event()
            except ResumeWebSocket:
                logger.info('Got a request to RESUME the websocket.')
                self.dispatch('disconnect')
                coro = cls.from_client(self, shard_id=self.shard_id, session=self.ws.session_id,
                                                    sequence=self.ws.sequence, resume=True)
                self.ws = await asyncio.wait_for(coro, timeout=180.0)



"""An experiment that allows you to register converters for classes,
as a helpful aid for readability when using typehints for custom converters.

You could also override base converters if you wanted to.

Example:
```py
class YertConverter(commands.Converter):
    async def convert(self, ctx, arg):
        return 'yert'

class Yert(object):
    ...

bot.converters[Yert] = YertConverter

@bot.command()
async def yert(ctx, yert: Yert):
    '''This will always send `yert`!'''
    await ctx.send(yert)
```
"""
from types import FunctionType

import nextcord
from nextcord.ext import commands
from nextcord.ext.commands import converter, Command

from ._common import _ALL

_BUILTINS = (
    bool,
    str,
    int,
    float,
)

_CONVERTERS = {
    # fmt: off
    nextcord.CategoryChannel: converter.CategoryChannelConverter,
    nextcord.Colour:          converter.ColourConverter,
    nextcord.Emoji:           converter.EmojiConverter,
    nextcord.Game:            converter.GameConverter,
    nextcord.Invite:          converter.InviteConverter,
    nextcord.Member:          converter.MemberConverter,
    nextcord.Message:         converter.MessageConverter,
    nextcord.PartialEmoji:    converter.PartialEmojiConverter,
    nextcord.Role:            converter.RoleConverter,
    nextcord.TextChannel:     converter.TextChannelConverter,
    nextcord.User:            converter.UserConverter,
    nextcord.VoiceChannel:    converter.VoiceChannelConverter,
    # fmt: on
}

_CONVERTERS.update({b: b for b in _BUILTINS})


class _ConverterDict(dict):
    """An easy way to register converters for classes.

    Can help for both linting and readability.
    """

    def __init__(self):
        super().__init__(_CONVERTERS)
        super().update(_ALL)

    def __setitem__(self, k, v):
        if not (isinstance(v, FunctionType) or issubclass(v, (*_BUILTINS, converter.Converter))):
            raise TypeError(
                "Excepted value of type 'Converter' or built-in, received %r" % v.__name__
            )
        super().__setitem__(k, v)

    def set(self, k, v):
        """Same as doing ``ConverterDict[Obj] = ObjConverter`` but fluid."""
        self.__setitem__(k, v)
        return self


_GLOBAL_CONVERTER_DICT = _ConverterDict()

commands.bot.BotBase.converters = _GLOBAL_CONVERTER_DICT

_old_actual_conversion = Command._actual_conversion


async def _actual_conversion(self, ctx, converter, argument, param):
    converter = _GLOBAL_CONVERTER_DICT.get(converter, converter)
    return await _old_actual_conversion(self, ctx, converter, argument, param)


Command._actual_conversion = _actual_conversion



import typing
from dataclasses import dataclass

from .flags import EventFlags, OpFlags
from .gateway import BunsenBurner
from .snoopy import Analyst
from .recorders.base import BaseRecorder
from .http import Beaker, Pipette


@dataclass
class Configuration:
    recorder: BaseRecorder
    event_flags: EventFlags = EventFlags.all()
    op_flags: OpFlags = OpFlags.all()
    gw_cls: BunsenBurner = BunsenBurner
    anal_cls: Analyst = Analyst
    httpclient_cls: Pipette = Pipette
    session_cls: Beaker = Beaker

Config = Configuration



import asyncio
import inspect
from collections import OrderedDict
from typing import (Any, Callable, Coroutine, Mapping, NoReturn, Optional,
                    OrderedDict, Union)

import nextcord
from nextcord.ext import commands
from nextcord.permissions import Permissions

from .constants import DEFAULT_TIMEOUT, EmojiType, log
from .exceptions import (CannotAddReactions, CannotEmbedLinks,
                         CannotReadMessageHistory, CannotSendMessages,
                         MenuError)
from .utils import Position, _cast_emoji


class Button:
    """Represents a reaction-style button for the :class:`Menu`.

    There are two ways to create this, the first being through explicitly
    creating this class and the second being through the decorator interface,
    :func:`button`.

    The action must have both a ``self`` and a ``payload`` parameter
    of type :class:`nextcord.RawReactionActionEvent`.

    Attributes
    ------------
    emoji: :class:`nextcord.PartialEmoji`
        The emoji to use as the button. Note that passing a string will
        transform it into a :class:`nextcord.PartialEmoji`.
    action
        A coroutine that is called when the button is pressed.
    skip_if: Optional[Callable[[:class:`Menu`], :class:`bool`]]
        A callable that detects whether it should be skipped.
        A skipped button does not show up in the reaction list
        and will not be processed.
    position: :class:`Position`
        The position the button should have in the initial order.
        Note that since nextcord does not actually maintain reaction
        order, this is a best effort attempt to have an order until
        the user restarts their client. Defaults to ``Position(0)``.
    lock: :class:`bool`
        Whether the button should lock all other buttons from being processed
        until this button is done. Defaults to ``True``.
    """
    __slots__ = ('emoji', '_action', '_skip_if', 'position', 'lock')

    def __init__(self, emoji: nextcord.PartialEmoji, action: Coroutine, *, skip_if: Optional[Callable[["Menu"], bool]] = None,
                 position: Optional[Position] = None, lock: Optional[bool] = True):

        self.emoji = _cast_emoji(emoji)
        self.action = action
        self.skip_if = skip_if
        self.position = position or Position(0)
        self.lock = lock

    @property
    def skip_if(self) -> Optional[Callable[["Menu"], bool]]:
        return self._skip_if

    @skip_if.setter
    def skip_if(self, value: Optional[Callable[["Menu"], bool]]):
        if value is None:
            self._skip_if = lambda _: False
            return

        try:
            menu_self = value.__self__
        except AttributeError:
            self._skip_if = value
        else:
            # Unfurl the method to not be bound
            if not isinstance(menu_self, Menu):
                raise TypeError(
                    'skip_if bound method must be from Menu not %r' % menu_self)

            self._skip_if = value.__func__

    @property
    def action(self) -> Coroutine:
        return self._action

    @action.setter
    def action(self, value: Coroutine):
        try:
            menu_self = value.__self__
        except AttributeError:
            pass
        else:
            # Unfurl the method to not be bound
            if not isinstance(menu_self, Menu):
                raise TypeError(
                    'action bound method must be from Menu not %r' % menu_self)

            value = value.__func__

        if not inspect.iscoroutinefunction(value):
            raise TypeError('action must be a coroutine not %r' % value)

        self._action = value

    def __call__(self, menu: "Menu", payload: nextcord.RawReactionActionEvent):
        if self.skip_if(menu):
            return
        return self._action(menu, payload)

    def __str__(self) -> str:
        return str(self.emoji)

    def is_valid(self, menu) -> bool:
        return not self.skip_if(menu)


def button(emoji: EmojiType, **kwargs):
    """Denotes a method to be button for the :class:`Menu`.

    The methods being wrapped must have both a ``self`` and a ``payload``
    parameter of type :class:`nextcord.RawReactionActionEvent`.

    The keyword arguments are forwarded to the :class:`Button` constructor.

    Example
    ---------

    .. code-block:: python3

        class MyMenu(Menu):
            async def send_initial_message(self, ctx, channel):
                return await channel.send(f'Hello {ctx.author}')

            @button('\\N{THUMBS UP SIGN}')
            async def on_thumbs_up(self, payload):
                await self.message.edit(content=f'Thanks {self.ctx.author}!')

            @button('\\N{THUMBS DOWN SIGN}')
            async def on_thumbs_down(self, payload):
                await self.message.edit(content=f"That's not nice {self.ctx.author}...")

    Parameters
    ------------
    emoji: Union[:class:`str`, :class:`nextcord.PartialEmoji`]
        The emoji to use for the button.
    """
    def decorator(func: Callable) -> Callable:
        func.__menu_button__ = _cast_emoji(emoji)
        func.__menu_button_kwargs__ = kwargs
        return func
    return decorator


class _MenuMeta(type):
    # noinspection PyMethodParameters
    @classmethod
    def __prepare__(cls, name, bases, **kwargs) -> OrderedDict:
        # This is needed to maintain member order for the buttons
        return OrderedDict()

    # noinspection PyMethodParameters
    def __new__(cls, name, bases, attrs, **kwargs) -> "_MenuMeta":
        buttons = []
        new_cls = super().__new__(cls, name, bases, attrs)

        inherit_buttons = kwargs.pop('inherit_buttons', True)
        if inherit_buttons:
            # walk MRO to get all buttons even in subclasses
            for base in reversed(new_cls.__mro__):
                for elem, value in base.__dict__.items():
                    try:
                        value.__menu_button__
                    except AttributeError:
                        continue
                    else:
                        buttons.append(value)
        else:
            for elem, value in attrs.items():
                try:
                    value.__menu_button__
                except AttributeError:
                    continue
                else:
                    buttons.append(value)

        new_cls.__inherit_buttons__ = inherit_buttons
        new_cls.__menu_buttons__ = buttons
        return new_cls

    def get_buttons(cls) -> OrderedDict:
        buttons = OrderedDict()
        for func in cls.__menu_buttons__:
            emoji = func.__menu_button__
            buttons[emoji] = Button(emoji, func, **func.__menu_button_kwargs__)
        return buttons


class Menu(metaclass=_MenuMeta):
    r"""An interface that allows handling menus by using reactions as buttons.

    Buttons should be marked with the :func:`button` decorator. Please note that
    this expects the methods to have a single parameter, the ``payload``. This
    ``payload`` is of type :class:`nextcord.RawReactionActionEvent`.

    Attributes
    ------------
    timeout: :class:`float`
        The timeout to wait between button inputs.
    delete_message_after: :class:`bool`
        Whether to delete the message after the menu interaction is done.
    clear_reactions_after: :class:`bool`
        Whether to clear reactions after the menu interaction is done.
        Note that :attr:`delete_message_after` takes priority over this attribute.
        If the bot does not have permissions to clear the reactions then it will
        delete the reactions one by one.
    check_embeds: :class:`bool`
        Whether to verify embed permissions as well.
    ctx: Optional[:class:`commands.Context`]
        The context that started this pagination session or ``None`` if it hasn't
        been started yet.
    bot: Optional[:class:`commands.Bot`]
        The bot that is running this pagination session or ``None`` if it hasn't
        been started yet.
    message: Optional[:class:`nextcord.Message`]
        The message that has been sent for handling the menu. This is the returned
        message of :meth:`send_initial_message`. You can set it in order to avoid
        calling :meth:`send_initial_message`\, if for example you have a pre-existing
        message you want to attach a menu to.
    """

    def __init__(self, *, timeout: float = DEFAULT_TIMEOUT, delete_message_after: bool = False,
                 clear_reactions_after: bool = False, check_embeds: bool = False, message: Optional[nextcord.Message] = None):

        self.timeout = timeout
        self.delete_message_after = delete_message_after
        self.clear_reactions_after = clear_reactions_after
        self.check_embeds = check_embeds
        self._can_remove_reactions = False
        self.__tasks = []
        self._running = True
        self.message = message
        self.ctx = None
        self.bot = None
        self._author_id = None
        self._buttons = self.__class__.get_buttons()
        self._lock = asyncio.Lock()
        self._event = asyncio.Event()

    @nextcord.utils.cached_property
    def buttons(self) -> Mapping[str, Button]:
        """Retrieves the buttons that are to be used for this menu session.

        Skipped buttons are not in the resulting dictionary.

        Returns
        ---------
        Mapping[:class:`str`, :class:`Button`]
            A mapping of button emoji to the actual button class.
        """
        buttons = sorted(self._buttons.values(), key=lambda b: b.position)
        return {
            button.emoji: button
            for button in buttons
            if button.is_valid(self)
        }

    def add_button(self, button: Button, *, react: bool = False):
        """|maybecoro|

        Adds a button to the list of buttons.

        If the menu has already been started then the button will
        not be added unless the ``react`` keyword-only argument is
        set to ``True``. Note that when this happens this function
        will need to be awaited.

        If a button with the same emoji is added then it is overridden.

        .. warning::

            If the menu has started and the reaction is added, the order
            property of the newly added button is ignored due to an API
            limitation with nextcord and the fact that reaction ordering
            is not guaranteed.

        Parameters
        ------------
        button: :class:`Button`
            The button to add.
        react: :class:`bool`
            Whether to add a reaction if the menu has been started.
            Note this turns the method into a coroutine.

        Raises
        ---------
        MenuError
            Tried to use ``react`` when the menu had not been started.
        nextcord.HTTPException
            Adding the reaction failed.
        """

        self._buttons[button.emoji] = button

        if react:
            if self.__tasks:
                async def wrapped():
                    # Add the reaction
                    try:
                        await self.message.add_reaction(button.emoji)
                    except nextcord.HTTPException:
                        raise
                    else:
                        # Update the cache to have the value
                        self.buttons[button.emoji] = button

                return wrapped()

            async def dummy():
                raise MenuError('Menu has not been started yet')
            return dummy()

    def remove_button(self, emoji: Union[Button, str], *, react: bool = False) -> Union[Coroutine[Any, Any, None], NoReturn]:
        """|maybecoro|

        Removes a button from the list of buttons.

        This operates similar to :meth:`add_button`.

        Parameters
        ------------
        emoji: Union[:class:`Button`, :class:`str`]
            The emoji or the button to remove.
        react: :class:`bool`
            Whether to remove the reaction if the menu has been started.
            Note this turns the method into a coroutine.

        Raises
        ---------
        MenuError
            Tried to use ``react`` when the menu had not been started.
        nextcord.HTTPException
            Removing the reaction failed.
        """

        if isinstance(emoji, Button):
            emoji = emoji.emoji
        else:
            emoji = _cast_emoji(emoji)

        self._buttons.pop(emoji, None)

        if react:
            if self.__tasks:
                async def wrapped():
                    # Remove the reaction from being processable
                    # Removing it from the cache first makes it so the check
                    # doesn't get triggered.
                    self.buttons.pop(emoji, None)
                    await self.message.remove_reaction(emoji, self.__me)
                return wrapped()

            async def dummy():
                raise MenuError('Menu has not been started yet')
            return dummy()

    def clear_buttons(self, *, react: bool = False) -> Union[Coroutine[Any, Any, None], NoReturn]:
        """|maybecoro|

        Removes all buttons from the list of buttons.

        If the menu has already been started then the buttons will
        not be removed unless the ``react`` keyword-only argument is
        set to ``True``. Note that when this happens this function
        will need to be awaited.

        Parameters
        ------------
        react: :class:`bool`
            Whether to clear the reactions if the menu has been started.
            Note this turns the method into a coroutine.

        Raises
        ---------
        MenuError
            Tried to use ``react`` when the menu had not been started.
        nextcord.HTTPException
            Clearing the reactions failed.
        """

        self._buttons.clear()

        if react:
            if self.__tasks:
                return self.clear

            async def dummy():
                raise MenuError('Menu has not been started yet')

            return dummy()

    def should_add_reactions(self) -> bool:
        """:class:`bool`: Whether to add reactions to this menu session."""
        return len(self.buttons) > 0

    def should_add_buttons(self) -> bool:
        """:class:`bool`: Whether to add button components to this menu session."""
        return hasattr(self, 'children') and len(self.children) > 0

    def should_add_reactions_or_buttons(self) -> bool:
        """:class:`bool`: Whether to add reactions or buttons to this menu session."""
        return self.should_add_reactions() or self.should_add_buttons()

    def _verify_permissions(self, ctx: commands.Context, channel: nextcord.abc.Messageable, permissions: Permissions):
        if not permissions.send_messages:
            raise CannotSendMessages()

        if self.check_embeds and not permissions.embed_links:
            raise CannotEmbedLinks()

        self._can_remove_reactions = permissions.manage_messages
        if self.should_add_reactions():
            if not permissions.add_reactions:
                raise CannotAddReactions()
            if not permissions.read_message_history:
                raise CannotReadMessageHistory()

    def reaction_check(self, payload: nextcord.RawReactionActionEvent) -> bool:
        """The function that is used to check whether the payload should be processed.
        This is passed to :meth:`nextcord.ext.commands.Bot.wait_for <Bot.wait_for>`.

        There should be no reason to override this function for most users.

        Parameters
        ------------
        payload: :class:`nextcord.RawReactionActionEvent`
            The payload to check.

        Returns
        ---------
        :class:`bool`
            Whether the payload should be processed.
        """
        if payload.message_id != self.message.id:
            return False
        if payload.user_id not in {self.bot.owner_id, self._author_id, *self.bot.owner_ids}:
            return False

        return payload.emoji in self.buttons

    async def _internal_loop(self):
        try:
            self.__timed_out = False
            loop = self.bot.loop
            # Ensure the name exists for the cancellation handling
            tasks = []
            while self._running:
                tasks = [
                    asyncio.ensure_future(self.bot.wait_for(
                        'raw_reaction_add', check=self.reaction_check)),
                    asyncio.ensure_future(self.bot.wait_for(
                        'raw_reaction_remove', check=self.reaction_check))
                ]
                done, pending = await asyncio.wait(tasks, timeout=self.timeout, return_when=asyncio.FIRST_COMPLETED)
                for task in pending:
                    task.cancel()

                if len(done) == 0:
                    raise asyncio.TimeoutError()

                # Exception will propagate if e.g. cancelled or timed out
                payload = done.pop().result()
                loop.create_task(self.update(payload))

                # NOTE: Removing the reaction ourselves after it's been done when
                # mixed with the checks above is incredibly racy.
                # There is no guarantee when the MESSAGE_REACTION_REMOVE event will
                # be called, and chances are when it does happen it'll always be
                # after the remove_reaction HTTP call has returned back to the caller
                # which means that the stuff above will catch the reaction that we
                # just removed.

                # For the future sake of myself and to save myself the hours in the future
                # consider this my warning.

        except asyncio.TimeoutError:
            self.__timed_out = True
        finally:
            self._event.set()

            # Cancel any outstanding tasks (if any)
            for task in tasks:
                task.cancel()

            try:
                await self.finalize(self.__timed_out)
            except Exception:
                pass
            finally:
                self.__timed_out = False

            # Can't do any requests if the bot is closed
            if self.bot.is_closed():
                return

            if self.delete_message_after:
                await self.message.delete()
            elif getattr(self, "clear_buttons_after", self.clear_reactions_after):
                await self.clear()
            elif getattr(self, "disable_buttons_after", None):
                await self.disable()

    async def update(self, payload: nextcord.RawReactionActionEvent):
        """|coro|

        Updates the menu after an event has been received.

        Parameters
        -----------
        payload: :class:`nextcord.RawReactionActionEvent`
            The reaction event that triggered this update.
        """
        button = self.buttons[payload.emoji]
        if not self._running:
            return

        try:
            if button.lock:
                async with self._lock:
                    if self._running:
                        await button(self, payload)
            else:
                await button(self, payload)
        except Exception as exc:
            await self.on_menu_button_error(exc)

    async def on_menu_button_error(self, exc: Exception):
        """|coro|

        Handles reporting of errors while updating the menu from events.
        The default behaviour is to log the exception.

        This may be overriden by subclasses.

        Parameters
        ----------
        exc: :class:`Exception`
            The exception which was raised during a menu update.
        """
        # some users may wish to take other actions during or beyond logging
        # which would require awaiting, such as stopping an erroring menu.
        log.exception("Unhandled exception during menu update.", exc_info=exc)

    async def start(self, ctx: commands.Context, *, channel: Optional[nextcord.abc.Messageable] = None, wait: bool = False):
        """|coro|

        Starts the interactive menu session.

        Parameters
        -----------
        ctx: :class:`Context`
            The invocation context to use.
        channel: :class:`nextcord.abc.Messageable`
            The messageable to send the message to. If not given
            then it defaults to the channel in the context.
        wait: :class:`bool`
            Whether to wait until the menu is completed before
            returning back to the caller.

        Raises
        -------
        MenuError
            An error happened when verifying permissions.
        nextcord.HTTPException
            Adding a reaction failed.
        """

        # Clear the buttons cache and re-compute if possible.
        try:
            del self.buttons
        except AttributeError:
            pass

        self.bot = bot = ctx.bot
        self.ctx = ctx
        self._author_id = ctx.author.id
        channel = channel or ctx.channel
        me = channel.guild.me if hasattr(channel, 'guild') else ctx.bot.user
        permissions = channel.permissions_for(me)
        self.__me = nextcord.Object(id=me.id)
        self._verify_permissions(ctx, channel, permissions)
        self._event.clear()
        msg = self.message
        if msg is None:
            self.message = msg = await self.send_initial_message(ctx, channel)

        if self.should_add_reactions_or_buttons():
            # Start the task first so we can listen to reactions before doing anything
            for task in self.__tasks:
                task.cancel()
            self.__tasks.clear()

            self._running = True
            self.__tasks.append(bot.loop.create_task(self._internal_loop()))

            async def add_reactions_task():
                for emoji in self.buttons:
                    await msg.add_reaction(emoji)
            self.__tasks.append(bot.loop.create_task(add_reactions_task()))

            if wait:
                await self._event.wait()

    async def finalize(self, timed_out: bool):
        """|coro|

        A coroutine that is called when the menu loop has completed
        its run. This is useful if some asynchronous clean-up is
        required after the fact.

        Parameters
        --------------
        timed_out: :class:`bool`
            Whether the menu completed due to timing out.
        """
        pass

    async def send_initial_message(self, ctx: commands.Context, channel: nextcord.abc.Messageable) -> nextcord.Message:
        """|coro|

        Sends the initial message for the menu session.

        This is internally assigned to the :attr:`message` attribute.

        Subclasses must implement this if they don't set the
        :attr:`message` attribute themselves before starting the
        menu via :meth:`start`.

        Parameters
        ------------
        ctx: :class:`Context`
            The invocation context to use.
        channel: :class:`nextcord.abc.Messageable`
            The messageable to send the message to.

        Returns
        --------
        :class:`nextcord.Message`
            The message that has been sent.
        """
        raise NotImplementedError

    async def clear(self):
        """|coro|

        Removes all reaction buttons in the menu.
        """
        # Wrap it in another block anyway just to ensure
        # nothing leaks out during clean-up
        try:
            # A fast path if we have permissions
            if self._can_remove_reactions:
                try:
                    del self.buttons
                except AttributeError:
                    pass
                finally:
                    await self.message.clear_reactions()
                return

            # Remove the cache (the next call will have the updated buttons)
            reactions = list(self.buttons.keys())
            try:
                del self.buttons
            except AttributeError:
                pass

            for reaction in reactions:
                try:
                    await self.message.remove_reaction(reaction, self.__me)
                except nextcord.HTTPException:
                    continue
        except Exception:
            pass

    def stop(self):
        """Stops the internal loop."""
        self._running = False
        for task in self.__tasks:
            task.cancel()
        self.__tasks.clear()


class ButtonMenu(Menu, nextcord.ui.View):
    r"""An interface that allows handling menus by using button interaction components.

    Buttons should be marked with the :func:`nextcord.ui.button` decorator. Please note that
    this expects the methods to have two parameters, the ``button`` and the ``interaction``.
    The ``button`` is of type :class:`nextcord.ui.Button`.
    The ``interaction`` is of type :class:`nextcord.Interaction`.

    Attributes
    ------------

    clear_buttons_after: :class:`bool`
        Whether to clear buttons after the menu interaction is done.
        Note that :attr:`delete_message_after` takes priority over this attribute.
    disable_buttons_after: :class:`bool`
        Whether to disable all buttons after the menu interaction is done.
        Note that :attr:`delete_message_after` and :attr:`clear_buttons_after` take priority over this attribute.
    """

    def __init__(self, timeout: float = DEFAULT_TIMEOUT, clear_buttons_after: bool = False,
                 disable_buttons_after: bool = False, *args, **kwargs):
        Menu.__init__(self, timeout=timeout, *args, **kwargs)
        nextcord.ui.View.__init__(self, timeout=timeout)
        self.clear_buttons_after = clear_buttons_after
        self.disable_buttons_after = disable_buttons_after

    async def _update_view(self):
        await self.message.edit(view=self)

    async def _set_all_disabled(self, disable: bool):
        """|coro|

        Enables or disable all :class:`nextcord.ui.Button` components in the menu.

        Parameters
        ------------
        disable: :class:`bool`
            Whether to disable or enable the buttons.
        """
        for child in self.children:
            if isinstance(child, nextcord.ui.Button):
                child.disabled = disable
        await self._update_view()

    async def enable(self):
        """|coro|

        Enables all :class:`nextcord.ui.Button` components in the menu.
        """
        await self._set_all_disabled(False)

    async def disable(self):
        """|coro|

        Disables all :class:`nextcord.ui.Button` components in the menu.
        """
        await self._set_all_disabled(True)

    async def clear(self):
        """|coro|

        Removes all :class:`nextcord.ui.Button` components in the menu.
        """
        self.clear_items()
        await self._update_view()

    def stop(self):
        """Stops the internal loop and view interactions."""
        # stop the menu loop
        Menu.stop(self)
        # stop view interactions
        nextcord.ui.View.stop(self)



from .constants import *
from .exceptions import *
from .menu_pages import *
from .menus import *
from .page_source import *
from .utils import *

# Needed for the setup.py script
__version__ = '1.3.2'



"""An experiment to allow for ``Message.delete`` to be silenced
of any exception. 

It uses a keyword argument called `silent`, and is by default
``False``.
"""

import nextcord


_old_delete = nextcord.Message.delete


async def delete(self, *, silent=False, **kwargs):
    try:
        await _old_delete(self, **kwargs)
    except Exception as e:
        if not silent:
            raise e


nextcord.Message.delete = delete



import sys


_ALL = {
    # This will be populated by loaded alternative converters at runtime
}


def py_allow(major: int, minor: int, micro: int) -> None:
    if sys.version_info < (major, minor, micro):
        raise RuntimeError(
            "This extension requires Python>={0}.{1}.{2}".format(major, minor, micro)
        )



import time
import logging
import typing
from functools import wraps

from aiohttp.typedefs import StrOrURL
from aiohttp.client_reqrep import ClientResponse

from .recorders.base import BaseRecorder
from .flags import EventFlags, OpFlags
from .op import OP_DICT, OpDetails

logger = logging.getLogger(__name__)


class ReadyTimer:
    start_time: float
    end_time: float

    def start(self):
        self.start_time = time.monotonic()
    
    def end(self) -> float:
        self.end_time = time.monotonic()
        return self.end_time - self.start_time

def common_table(table, *, unknown=False):
    def wrapper(func):
        @wraps(func)
        async def wrapped(self, *args, **kwargs) -> None:
            save = getattr(self.recorder, 'save_{}'.format(table))
            args = await func(self, *args, **kwargs)
            if args is None:
                return
            
            if unknown:
                await save(*args, **kwargs, unknown=unknown)
            else:
                await save(*args, **kwargs)
        return wrapped
    return wrapper


class Analyst:
    ready_timer = ReadyTimer()

    def __init__(self, recorder: BaseRecorder, event_flags: EventFlags, op_flags: OpFlags):
        self.recorder = recorder
        self.event_flags = event_flags,
        self.op_flags = op_flags

    async def log(self, event_name, *args, **kwargs):
        handler = getattr(self, 'on_{}'.format(event_name), None)
        try:
            payload: dict = args[0]
        except IndexError:
            return
        
        if type(payload) is bytes:
            return # ping
        
        if handler is None:
            return await self.on_unknown_event(event_name, payload)
        
        if type(payload) is dict:
            await self.recorder.save_events(event_name, payload)
        # TODO: serialize dataclass objects
        await handler(*args, **kwargs)
    
    @common_table('events', unknown=True)
    async def on_unknown_event(self, event_name: str, payload: dict) -> typing.Tuple[str, dict]:
        if type(payload) is dict:
            return event_name, payload

    # TODO: log events

    @common_table('packets')
    async def on_socket_response(self, payload: dict) -> typing.Optional[typing.Tuple[int, OpDetails]]:
        op_code = payload['op']
        
        handler = getattr(self, 'on_socket_op_{}'.format(op_code), None)
        
        details = OpDetails(inbound=True, payload=payload)
        if handler:
            details = await handler(payload)
        
        should_log = OP_DICT[op_code](self.op_flags)
        if should_log:
            return op_code, details
    
    @common_table('packets')
    async def on_socket_send(self, payload: dict) -> typing.Optional[typing.Tuple[int, OpDetails]]:
        op_code = payload['op']

        # TODO: propagate

        details = OpDetails(inbound=False, payload=payload)

        should_log = OP_DICT[op_code](self.op_flags)
        if should_log:
            return op_code, details
    
    async def on_socket_op_0(self, payload: dict) -> OpDetails:
        event_name = payload['t']

        handler = getattr(self, 'on_socket_{}'.format(event_name), None)

        if handler:
            await handler(payload)

        return OpDetails(inbound=True, event_name=event_name, payload=payload)
    
    async def on_socket_READY(self, payload: dict) -> None:
        duration = self.ready_timer.end()
        logger.debug("Received READY event {:,.2f} milliseconds after connection.".format(duration * 1000))
    
    async def on_socket_RESUME(self, payload: dict) -> None:
        duration = self.ready_timer.end()
        logger.debug("Received RESUME event {:,.2f} milliseconds after reconnection.".format(duration * 1000))

    @common_table('requests')
    async def http_request(self, method: str, str_or_url: StrOrURL, **kwargs: dict) -> typing.Tuple[typing.Any, ...]:
        return method, str_or_url, kwargs
    
    @common_table('requests')
    async def http_request_error(self, method: str, str_or_url: StrOrURL, **kwargs: dict) -> typing.Tuple[typing.Any, ...]:
        return method, str_or_url, kwargs
